# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import jax.numpy as jnp
from jaxtyping import Float
import numpyro.distributions as npd

from gpjax.kernels.stationary.base import StationaryKernel
from gpjax.kernels.stationary.utils import (
    build_student_t_distribution,
    euclidean_distance,
)
from gpjax.typing import Array


class Matern52(StationaryKernel):
    r"""The Matérn kernel with smoothness parameter fixed at 2.5.


    Computes the covariance for pairs of inputs $(x, y)$ with
    lengthscale parameter $\ell$ and variance $\sigma^2$.

    $$
    k(x, y) = \sigma^2 \exp \Bigg(1+ \frac{\sqrt{5}\lvert x-y \rvert}{\ell^2} + \frac{5\lvert x - y \rvert^2}{3\ell^2} \Bigg)\exp\Bigg(-\frac{\sqrt{5}\lvert x-y\rvert}{\ell^2} \Bigg)
    $$
    """

    name: str = "Matérn52"

    def __call__(
        self, x: Float[Array, " D"], y: Float[Array, " D"]
    ) -> Float[Array, ""]:
        x = self.slice_input(x) / self.lengthscale.value
        y = self.slice_input(y) / self.lengthscale.value
        tau = euclidean_distance(x, y)
        K = (
            self.variance.value
            * (1.0 + jnp.sqrt(5.0) * tau + 5.0 / 3.0 * jnp.square(tau))
            * jnp.exp(-jnp.sqrt(5.0) * tau)
        )
        return K.squeeze()

    @property
    def spectral_density(self) -> npd.StudentT:
        return build_student_t_distribution(nu=5)



import pathlib

from mktestdocs import check_md_file
import pytest


# Ensure that code chunks within any markdown files execute without error
@pytest.mark.parametrize("fpath", pathlib.Path("gpjax/").glob("*.md"), ids=str)
def test_source_good(fpath):
    check_md_file(fpath=fpath, memory=True)


@pytest.mark.parametrize("fpath", pathlib.Path("docs").glob("*.md"), ids=str)
def test_docs_good(fpath):
    check_md_file(fpath=fpath, memory=True)


@pytest.mark.parametrize("fpath", pathlib.Path(".").glob("*.md"), ids=str)
def test_root_good(fpath):
    check_md_file(fpath=fpath, memory=True)



# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     custom_cell_magics: kql
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.7
#   kernelspec:
#     display_name: gpjax_beartype
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Sparse Stochastic Variational Inference
#
# In this notebook we demonstrate how to implement sparse variational Gaussian
# processes (SVGPs) of
# <strong data-cite="hensman2015gaussian">Hensman et al. (2015)</strong>. In
# particular, this approximation framework provides a tractable option for working with
# non-conjugate Gaussian processes with more than ~5000 data points. However, for
# conjugate models of less than 5000 data points, we recommend using the marginal
# log-likelihood approach presented in the
# [regression notebook](https://docs.jaxgaussianprocesses.com/_examples/regression/).
# Though we illustrate SVGPs here with a conjugate regression example, the same GPJax
# code works for general likelihoods, such as a Bernoulli for classification.


# %%
# Enable Float64 for more stable matrix inversions.
from jax import config
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import install_import_hook
import matplotlib as mpl
import matplotlib.pyplot as plt
import optax as ox

from examples.utils import use_mpl_style

config.update("jax_enable_x64", True)


with install_import_hook("gpjax", "beartype.beartype"):
    import gpjax as gpx
    import gpjax.kernels as jk

key = jr.key(123)

# set the default style for plotting
use_mpl_style()
cols = mpl.rcParams["axes.prop_cycle"].by_key()["color"]


# %% [markdown]
# ## Dataset
#
# With the necessary modules imported, we simulate a dataset
# $\mathcal{D} = (\boldsymbol{x}, \boldsymbol{y}) = \{(x_i, y_i)\}_{i=1}^{5000}$
# with inputs $\boldsymbol{x}$ sampled uniformly on $(-5, 5)$ and corresponding binary outputs
#
# $$\boldsymbol{y} \sim \mathcal{N} \left(\sin(4 * \boldsymbol{x}) + \sin(2 * \boldsymbol{x}), \textbf{I} * (0.2)^{2} \right).$$
#
# We store our data $\mathcal{D}$ as a GPJax `Dataset` and create test inputs for later.

# %%
n = 50000
noise = 0.2

key, subkey = jr.split(key)
x = jr.uniform(key=key, minval=-5.0, maxval=5.0, shape=(n,)).reshape(-1, 1)
f = lambda x: jnp.sin(4 * x) + jnp.cos(2 * x)
signal = f(x)
y = signal + jr.normal(subkey, shape=signal.shape) * noise
D = gpx.Dataset(X=x, y=y)

xtest = jnp.linspace(-5.5, 5.5, 500).reshape(-1, 1)

# %% [markdown]
# ## Sparse GPs via inducing inputs
#
# Despite their endowment with elegant theoretical properties, GPs are burdened with
# prohibitive $\mathcal{O}(n^3)$ inference and $\mathcal{O}(n^2)$ memory costs in the
# number of data points $n$ due to the necessity of computing inverses and determinants
# of the kernel Gram matrix $\mathbf{K}_{\boldsymbol{x}\boldsymbol{x}}$ during inference
# and hyperparameter learning.
# Sparse GPs seek to resolve tractability through low-rank approximations.
#
# Their name originates with the idea of using subsets of the data to approximate the
# kernel matrix, with _sparseness_ occurring through the selection of the data points.
# Given inputs $\boldsymbol{x}$ and outputs $\boldsymbol{y}$ the task was to select an
# $m<n$ lower-dimensional dataset $(\boldsymbol{z},\boldsymbol{\tilde{y}}) \subset (\boldsymbol{x}, \boldsymbol{y})$
# to train a Gaussian process on instead.
# By generalising the set of selected points $\boldsymbol{z}$, known as
# _inducing inputs_, to remove the restriction of being part of the dataset,
# we can arrive at a flexible low-rank approximation framework of the model using
# functions of $\mathbf{K}_{\boldsymbol{z}\boldsymbol{z}}$ to replace the true
# covariance matrix $\mathbf{K}_{\boldsymbol{x}\boldsymbol{x}}$ at significantly
# lower costs. For example, <strong data-cite="quinonero-candela2005gaussian"></strong>
# review many popular approximation schemes in this vein. However, because the model
# and the approximation are intertwined, assigning performance and faults to one or the
# other becomes tricky.
#
# On the other hand, sparse variational Gaussian processes (SVGPs)
# [approximate the posterior, not the model](https://www.secondmind.ai/labs/sparse-gps-approximate-the-posterior-not-the-model/).
# These provide a low-rank approximation scheme via variational inference. Here we
# posit a family of densities parameterised by "variational parameters".
# We then seek to find the closest family member to the posterior by minimising the
# Kullback-Leibler divergence over the variational parameters.
# The fitted variational density then serves as a proxy for the exact posterior.
# This procedure makes variational methods efficiently solvable via off-the-shelf
# optimisation techniques whilst retaining the true-underlying model.
# Furthermore, SVGPs offer further cost reductions with mini-batch stochastic gradient
# descent  <strong data-cite="hensman2013gaussian"></strong> and address non-conjugacy
# <strong data-cite="hensman2015gaussian"></strong>.
# We show a cost comparison between the approaches below, where $b$ is the mini-batch
# size.
#
# |    | GPs | sparse GPs | SVGP |
# | -- | -- | -- | -- |
# | Inference cost | $\mathcal{O}(n^3)$ | $\mathcal{O}(n m^2)$ | $\mathcal{O}(b m^2 + m^3)$  |
# | Memory cost    | $\mathcal{O}(n^2)$ | $\mathcal{O}(n m)$ | $\mathcal{O}(b m + m^2)$ |
#
# To apply SVGP inference to our dataset, we begin by initialising $m = 50$ equally
# spaced inducing inputs $\boldsymbol{z}$ across our observed data's support. These
# are depicted below via horizontal black lines.

# %%
z = jnp.linspace(-5.0, 5.0, 50).reshape(-1, 1)

fig, ax = plt.subplots()
ax.vlines(
    z,
    ymin=y.min(),
    ymax=y.max(),
    alpha=0.3,
    linewidth=1,
    label="Inducing point",
    color=cols[2],
)
ax.scatter(x, y, alpha=0.2, color=cols[0], label="Observations")
ax.plot(xtest, f(xtest), color=cols[1], label="Latent function")
ax.legend()
ax.set(xlabel=r"$x$", ylabel=r"$f(x)$")

# %% [markdown]
# The inducing inputs will summarise our dataset, and since they are treated as
# variational parameters, their locations will be optimised. The next step to SVGP is
# to define a variational family.

# %% [markdown]
# ## Defining the variational process
#
# We begin by considering the form of the posterior distribution for all function
# values $f(\cdot)$
#
# \begin{align}
# p(f(\cdot) | \mathcal{D}) = \int p(f(\cdot)|f(\boldsymbol{x})) p(f(\boldsymbol{x})|\mathcal{D}) \text{d}f(\boldsymbol{x}). \qquad (\dagger)
# \end{align}
#
# To arrive at an approximation framework, we assume some redundancy in the data.
# Instead of predicting $f(\cdot)$ with function values at the datapoints
# $f(\boldsymbol{x})$, we assume this can be achieved with only function values at
# $m$ inducing inputs $\boldsymbol{z}$
#
# $$ p(f(\cdot) | \mathcal{D}) \approx \int p(f(\cdot)|f(\boldsymbol{z})) p(f(\boldsymbol{z})|\mathcal{D}) \text{d}f(\boldsymbol{z}). \qquad (\star) $$
#
# This lower dimensional integral results in computational savings in the model's
# predictive component from $p(f(\cdot)|f(\boldsymbol{x}))$ to
# $p(f(\cdot)|f(\boldsymbol{z}))$ where inverting
# $\mathbf{K}_{\boldsymbol{x}\boldsymbol{x}}$ is replaced with inverting
# $\mathbf{K}_{\boldsymbol{z}\boldsymbol{z}}$.
# However, since we did not observe our data $\mathcal{D}$ at $\boldsymbol{z}$ we ask,
# what exactly is the posterior $p(f(\boldsymbol{z})|\mathcal{D})$?
#
# Notice this is simply obtained by substituting $\boldsymbol{z}$ into $(\dagger)$,
# but we arrive back at square one with computing the expensive integral. To side-step
# this, we consider replacing $p(f(\boldsymbol{z})|\mathcal{D})$ in $(\star)$ with a
# cheap-to-compute approximate distribution $q(f(\boldsymbol{z}))$
#
#   $$ q(f(\cdot)) = \int p(f(\cdot)|f(\boldsymbol{z})) q(f(\boldsymbol{z})) \text{d}f(\boldsymbol{z}). \qquad (\times) $$
#
# To measure the quality of the approximation, we consider the Kullback-Leibler
# divergence $\operatorname{KL}(\cdot || \cdot)$ from our approximate process
# $q(f(\cdot))$ to the true process $p(f(\cdot)|\mathcal{D})$. By parametrising
# $q(f(\boldsymbol{z}))$ over a variational family of distributions, we can optimise
# Kullback-Leibler divergence with respect to the variational parameters. Moreover,
# since inducing input locations $\boldsymbol{z}$ augment the model, they themselves
# can be treated as variational parameters without altering the true underlying model
# $p(f(\boldsymbol{z})|\mathcal{D})$. This is exactly what gives SVGPs great
# flexibility whilst retaining robustness to overfitting.
#
# It is popular to elect a Gaussian variational distribution
# $q(f(\boldsymbol{z})) = \mathcal{N}(f(\boldsymbol{z}); \mathbf{m}, \mathbf{S})$
# with parameters $\{\boldsymbol{z}, \mathbf{m}, \mathbf{S}\}$, since conjugacy is
# provided between $q(f(\boldsymbol{z}))$ and $p(f(\cdot)|f(\boldsymbol{z}))$ so that
# the resulting variational process $q(f(\cdot))$ is a GP. We can implement this in
# GPJax by the following.

# %%
meanf = gpx.mean_functions.Zero()
likelihood = gpx.likelihoods.Gaussian(num_datapoints=n)
kernel = jk.RBF()  # 1-dimensional inputs
prior = gpx.gps.Prior(mean_function=meanf, kernel=kernel)
p = prior * likelihood
q = gpx.variational_families.VariationalGaussian(posterior=p, inducing_inputs=z)

# %% [markdown]
# Here, the variational process $q(\cdot)$ depends on the prior through
# $p(f(\cdot)|f(\boldsymbol{z}))$ in $(\times)$.

# %% [markdown]
# ## Inference
#
# ### Evidence lower bound
#
# With our model defined, we seek to infer the optimal inducing inputs
# $\boldsymbol{z}$, variational mean $\mathbf{m}$ and covariance
# $\mathbf{S}$ that define our approximate posterior. To achieve this, we maximise the
# evidence lower bound (ELBO) with respect to
# $\{\boldsymbol{z}, \mathbf{m}, \mathbf{S} \}$, a proxy for minimising the
# Kullback-Leibler divergence. Moreover, as hinted by its name, the ELBO is a lower
# bound to the marginal log-likelihood, providing a tractable objective to optimise the
# model's hyperparameters akin to the conjugate setting. For further details on this,
# see Sections 3.1 and 4.1 of the excellent review paper
# <strong data-cite="leibfried2020tutorial"></strong>.
#

# %% [markdown]
# ### Mini-batching
#
# Despite introducing inducing inputs into our model, inference can still be
# intractable with large datasets. To circumvent this, optimisation can be done using
# stochastic mini-batches.

# %%
schedule = ox.warmup_cosine_decay_schedule(
    init_value=0.0,
    peak_value=0.02,
    warmup_steps=75,
    decay_steps=2000,
    end_value=0.001,
)

opt_posterior, history = gpx.fit(
    model=q,
    # we are minimizing the elbo so we negate it
    objective=lambda p, d: -gpx.objectives.elbo(p, d),
    train_data=D,
    optim=ox.adam(learning_rate=schedule),
    num_iters=3000,
    key=jr.key(42),
    batch_size=128,
)
# %% [markdown]
# ## Predictions
#
# With optimisation complete, we can use our inferred parameter set to make
# predictions at novel inputs akin
# to all other models within GPJax on our variational process object $q(\cdot)$ (for
# example, see the
# [regression notebook](https://docs.jaxgaussianprocesses.com/_examples/regression/)).

# %%
latent_dist = opt_posterior(xtest)
predictive_dist = opt_posterior.posterior.likelihood(latent_dist)

meanf = predictive_dist.mean
sigma = jnp.sqrt(predictive_dist.variance)

fig, ax = plt.subplots()
ax.scatter(x, y, alpha=0.15, label="Training Data", color=cols[0])
ax.plot(xtest, meanf, label="Posterior mean", color=cols[1])
ax.fill_between(
    xtest.flatten(),
    meanf - 2 * sigma,
    meanf + 2 * sigma,
    alpha=0.3,
    color=cols[1],
    label="Two sigma",
)
ax.vlines(
    opt_posterior.inducing_inputs.value,
    ymin=y.min(),
    ymax=y.max(),
    alpha=0.3,
    linewidth=1,
    label="Inducing point",
    color=cols[2],
)
ax.legend()

# %% [markdown]
# ## System configuration

# %%
# %reload_ext watermark
# %watermark -n -u -v -iv -w -a 'Thomas Pinder, Daniel Dodd & Zeel B Patel'



""" Convert python files in "examples" directory to markdown files using jupytext and nbconvert.

There's only a minor inconvenience with how supporting files are handled by nbconvert,
see https://github.com/jupyter/nbconvert/issues/1164. But these will be under a private
directory `_examples` in the docs folder, so it's not a big deal.

"""
from argparse import ArgumentParser
from pathlib import Path
import subprocess
from concurrent.futures import ThreadPoolExecutor, as_completed
import shutil

EXCLUDE = ["utils.py"]


def process_file(file: Path, out_file: Path | None = None, execute: bool = False):
    """Converts a python file to markdown using jupytext and nbconvert."""

    out_dir = out_file.parent
    command = f"cd {out_dir.as_posix()} && "

    out_file = out_file.relative_to(out_dir).as_posix()

    if execute:
        command += f"jupytext --to ipynb {file} --output - "
        command += (
            f"| jupyter nbconvert --to markdown --execute --stdin --output {out_file}"
        )
    else:
        command += f"jupytext --to markdown {file} --output {out_file}"

    subprocess.run(command, shell=True, check=False)


def is_modified(file: Path, out_file: Path):
    """Check if the output file is older than the input file."""
    return out_file.exists() and out_file.stat().st_mtime < file.stat().st_mtime


def main(args):
    # project root directory
    wdir = Path(__file__).parents[2]

    # output directory
    out_dir: Path = args.outdir
    out_dir.mkdir(exist_ok=True, parents=True)

    # copy directories in "examples" to output directory
    for dir in wdir.glob("examples/*"):
        if dir.is_dir():
            (out_dir / dir.name).mkdir(exist_ok=True, parents=True)
            for file in dir.glob("*"):
                # copy, not move!
                shutil.copy(file, out_dir / dir.name / file.name)

    # list of files to be processed
    files = [f for f in wdir.glob("examples/*.py") if f.name not in EXCLUDE]

    # process only modified files
    if args.only_modified:
        files = [f for f in files if is_modified(f, out_dir / f"{f.stem}.md")]

    print(files)

    # process files in parallel
    if args.parallel:
        with ThreadPoolExecutor(max_workers=args.max_workers) as executor:
            futures = []
            for file in files:
                out_file = out_dir / f"{file.stem}.md"
                futures.append(
                    executor.submit(
                        process_file, file, out_file=out_file, execute=args.execute
                    )
                )

            for future in as_completed(futures):
                try:
                    future.result()
                except Exception as e:
                    print(f"Error processing file: {e}")
    else:
        for file in files:
            out_file = out_dir / f"{file.stem}.md"
            process_file(file, out_file=out_file, execute=args.execute)


if __name__ == "__main__":
    project_root = Path(__file__).parents[2]

    parser = ArgumentParser()
    parser.add_argument("--max_workers", type=int, default=4)
    parser.add_argument("--execute", action="store_true")
    parser.add_argument("--only_modified", action="store_true")
    parser.add_argument(
        "--outdir", type=Path, default=project_root / "docs" / "_examples"
    )
    parser.add_argument("--parallel", type=bool, default=False)
    args = parser.parse_args()

    main(args)



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     custom_cell_magics: kql
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.11.2
#   kernelspec:
#     display_name: gpjax_baselines
#     language: python
#     name: python3
# ---

# %%
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib import patches

plt.style.use(
    "https://raw.githubusercontent.com/JaxGaussianProcesses/GPJax/main/docs/examples/gpjax.mplstyle"
)
cols = mpl.rcParams["axes.prop_cycle"].by_key()["color"]

# %%
fig, ax = plt.subplots()
ax.axhline(y=0.25, color=cols[0], linewidth=1.5)

xs = [0.02, 0.06, 0.1, 0.17]
ys = np.ones_like(xs) * 0.25

ax.scatter(xs, ys, color=cols[1], marker="o", s=100, zorder=2)

for idx, x in enumerate(xs):
    ax.annotate(
        text=f"$\\ell_{{t-{idx+1}}}$",
        xy=(x, 0.25),
        xytext=(x + 0.01, 0.275),
        ha="center",
        va="bottom",
    )


style = "Simple, tail_width=0.5, head_width=4, head_length=8"
kw = dict(arrowstyle=style, color="k")

for i in range(len(xs) - 1):
    a = patches.FancyArrowPatch(
        (xs[i + 1], 0.25), (xs[i], 0.25), connectionstyle="arc3,rad=-.5", **kw
    )
    ax.add_patch(a)


ax.scatter(-0.03, 0.25, color=cols[1], marker="x", s=100, linewidth=5, zorder=2)

a = patches.FancyArrowPatch(
    (xs[0], 0.25), (-0.03, 0.25), connectionstyle="arc3,rad=-.5", **kw
)
ax.add_patch(a)

ax.axvline(x=0, color="black", linewidth=0.5, linestyle="-.")
ax.get_yaxis().set_visible(False)
ax.spines["left"].set_visible(False)
ax.set_ylim(0.0, 0.5)
ax.set_xlim(-0.07, 0.25)
plt.savefig("../_static/step_size_figure.png", bbox_inches="tight")

# %%
import numpyro.distributions.transforms as npt

bij = npt.ExpTransform()

x = np.linspace(0.05, 3.0, 6)
y = np.asarray(bij.inv(x))
lval = 0.5
rval = 0.52

fig, ax = plt.subplots()
ax.scatter(x, np.ones_like(x) * lval, s=100, label="Constrained value")
ax.scatter(y, np.ones_like(y) * rval, marker="o", s=100, label="Unconstrained value")

style = "Simple, tail_width=0.25, head_width=2, head_length=8"
for i in range(len(x)):
    if i % 2 != 0:
        a = patches.FancyArrowPatch(
            (x[i], lval), (y[i], rval), connectionstyle="arc3,rad=-.15", **kw
        )
    # a = patches.Arrow(lval, x[i], rval-lval, y[i]-x[i], width=0.05, color='k')
    else:
        a = patches.FancyArrowPatch(
            (x[i], lval), (y[i], rval), connectionstyle="arc3,rad=.005", **kw
        )
    ax.add_patch(a)

ax.get_yaxis().set_visible(False)
ax.spines["left"].set_visible(False)
ax.legend(loc="best")
# ax.set_ylim(0.1, 0.32)
plt.savefig("../_static/bijector_figure.svg", bbox_inches="tight")

# %%
np.log(0.05)

# %%
print(x)



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import jax.numpy as jnp
from jaxtyping import Float
import numpyro.distributions as npd

from gpjax.typing import (
    Array,
    ScalarFloat,
)


def build_student_t_distribution(nu: int) -> npd.StudentT:
    r"""Build a Student's t distribution with a fixed smoothness parameter.

    For a fixed half-integer smoothness parameter, compute the spectral density of a
    Matérn kernel; a Student's t distribution.

    Args:
        nu (int): The smoothness parameter of the Matérn kernel.

    Returns
    -------
        tfp.Distribution: A Student's t distribution with the same smoothness parameter.
    """
    dist = npd.StudentT(df=nu, loc=0.0, scale=1.0)
    return dist


def squared_distance(x: Float[Array, " D"], y: Float[Array, " D"]) -> ScalarFloat:
    r"""Compute the squared distance between a pair of inputs.

    Args:
        x (Float[Array, " D"]): First input.
        y (Float[Array, " D"]): Second input.

    Returns
    -------
        ScalarFloat: The squared distance between the inputs.
    """
    return jnp.sum((x - y) ** 2)


def euclidean_distance(x: Float[Array, " D"], y: Float[Array, " D"]) -> ScalarFloat:
    r"""Compute the euclidean distance between a pair of inputs.

    Args:
        x (Float[Array, " D"]): First input.
        y (Float[Array, " D"]): Second input.

    Returns
    -------
        ScalarFloat: The euclidean distance between the inputs.
    """
    return jnp.sqrt(jnp.maximum(squared_distance(x, y), 1e-36))



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     custom_cell_magics: kql
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.7
#   kernelspec:
#     display_name: gpjax
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Gaussian Processes for Vector Fields and Ocean Current Modelling
#
# In this notebook, we use Gaussian processes to learn vector-valued functions. We will
# be recreating the results by [Berlinghieri et al.
# (2023)](https://arxiv.org/pdf/2302.10364.pdf) by an application to real-world ocean
# surface velocity data, collected via surface drifters.
#
# Surface drifters are measurement devices that measure the dynamics and circulation
# patterns of the world's oceans. Studying and predicting ocean currents are important
# to climate research, for example, forecasting and predicting oil spills, oceanographic
# surveying of eddies and upwelling, or providing information on the distribution of
# biomass in ecosystems. We will be using the [Gulf Drifters Open
# dataset](https://zenodo.org/record/4421585), which contains all publicly available
# surface drifter trajectories from the Gulf of Mexico spanning 28 years.
#

# %%
from dataclasses import (
    dataclass,
    field,
)

from jax import (
    config,
    hessian,
)
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import (
    Array,
    Float,
    install_import_hook,
)
from matplotlib import rcParams
import matplotlib.pyplot as plt
import numpyro.distributions as npd
import pandas as pd

from examples.utils import use_mpl_style
from gpjax.kernels.computations import DenseKernelComputation

config.update("jax_enable_x64", True)


with install_import_hook("gpjax", "beartype.beartype"):
    import gpjax as gpx


# set the default style for plotting
use_mpl_style()

key = jr.key(42)

colors = rcParams["axes.prop_cycle"].by_key()["color"]


# %% [markdown]
# ## Data loading and preprocessing
# The real dataset has been binned into an $N=34\times16$ grid, equally spaced over the
# longitude-latitude interval $[-90.8,-83.8] \times [24.0,27.5]$. Each bin has a size
# $\approx 0.21\times0.21$, and contains the average velocity across all measurements
# that fall inside it.
#
# We will call this binned ocean data the ground truth, and label it with the vector
# field $$ \mathbf{F} \equiv \mathbf{F}(\mathbf{x}), $$ where $\mathbf{x} =
# (x^{(0)}$,$x^{(1)})^\text{T}$, with a vector basis in the standard Cartesian
# directions (dimensions will be indicated by superscripts).
#
# We shall label the ground truth $D_0=\left\{ \left(\mathbf{x}_{0,i} , \mathbf{y}_{0,i}
# \right)\right\}_{i=1}^N$, where $\mathbf{y}_{0,i}$ is the 2-dimensional velocity
# vector at the $i$-th location, $\mathbf{x}_{0,i}$. The training dataset contains
# simulated measurements from ocean drifters $D_T=\left\{\left(\mathbf{x}_{T,i},
# \mathbf{y}_{T,i} \right)\right\}_{i=1}^{N_T}$, $N_T = 20$ in this case (the subscripts
# indicate the ground truth and the simulated measurements respectively).
#


# %%
# function to place data from csv into correct array shape
def prepare_data(df):
    pos = jnp.array([df["lon"], df["lat"]])
    vel = jnp.array([df["ubar"], df["vbar"]])
    # extract shape stored as 'metadata' in the test data
    try:
        shape = (int(df["shape"][1]), int(df["shape"][0]))  # shape = (34,16)
        return pos, vel, shape
    except KeyError:
        return pos, vel


# loading in data

gulf_data_train = pd.read_csv(
    "https://raw.githubusercontent.com/JaxGaussianProcesses/static/main/data/gulfdata_train.csv"
)
gulf_data_test = pd.read_csv(
    "https://raw.githubusercontent.com/JaxGaussianProcesses/static/main/data/gulfdata_test.csv"
)


pos_test, vel_test, shape = prepare_data(gulf_data_test)
pos_train, vel_train = prepare_data(gulf_data_train)

fig, ax = plt.subplots(1, 1, figsize=(6, 3))
ax.quiver(
    pos_test[0],
    pos_test[1],
    vel_test[0],
    vel_test[1],
    color=colors[0],
    label="Ocean Current",
    angles="xy",
    scale=10,
)
ax.quiver(
    pos_train[0],
    pos_train[1],
    vel_train[0],
    vel_train[1],
    color=colors[1],
    alpha=0.7,
    label="Drifter",
    angles="xy",
    scale=10,
)

ax.set(
    xlabel="Longitude",
    ylabel="Latitude",
)
ax.legend(
    framealpha=0.0,
    ncols=2,
    fontsize="medium",
    bbox_to_anchor=(0.5, -0.3),
    loc="lower center",
)

# %% [markdown]
# ## Problem Setting
# We aim to obtain estimates for $\mathbf{F}$ at the set of points $\left\{
# \mathbf{x}_{0,i} \right\}_{i=1}^N$ using Gaussian processes, followed by a comparison
# of the latent model to the ground truth $D_0$. Note that $D_0$ is not passed into any
# functions used  by GPJax, and is only used to compare against the two GP models at the
# end of the notebook.
#
# Since $\mathbf{F}$ is a vector-valued function, we require GPs that can directly learn
# vector-valued functions[<sup>1</sup>](#fn1). To implement this in GPJax, the problem
# can be changed to learn a scalar-valued function by 'massaging' the data into a
# $2N\times2N$ problem, such that each dimension of our GP is associated with a
# *component* of $\mathbf{y}_{T,i}$.
#
# For a particular measurement $\mathbf{y}$ (training or testing) at location
# $\mathbf{x}$, the components $(y^{(0)}, y^{(1)})$ are described by the latent vector
# field $\mathbf{F}$, such that
#
# $$ \mathbf{y} = \mathbf{F}(\mathbf{x}) = \left(\begin{array}{l}
# f^{(0)}\left(\mathbf{x}\right) \\
# f^{(1)}\left(\mathbf{x}\right) \end{array}\right), $$
#
# where each $f^{(z)}\left(\mathbf{x}\right), z \in \{0,1\}$ is a scalar-valued
# function.
#
# Now consider the scalar-valued function $g: \mathbb{R}^2 \times\{0,1\} \rightarrow
# \mathbb{R}$, such that
#
# $$ g \left(\mathbf{x} , 0 \right) = f^{(0)} ( \mathbf{x} ), \text{and } g \left(
# \mathbf{x}, 1 \right)=f^{(1)}\left(\mathbf{x}\right).  $$
#
# We have increased the input dimension by 1, from the 2D $\mathbf{x}$ to the 3D
# $\mathbf{X} = \left(\mathbf{x}, 0\right)$ or $\mathbf{X} = \left(\mathbf{x},
# 1\right)$.
#
# By choosing the value of the third dimension, 0 or 1, we may now incorporate this
# information into the computation of the kernel.  We therefore make new 3D datasets
# $D_{T,3D} = \left\{\left( \mathbf{X}_{T,i},\mathbf{Y}_{T,i} \right) \right\}
# _{i=0}^{2N_T}$ and $D_{0,3D} = \left\{\left( \mathbf{X}_{0,i},\mathbf{Y}_{0,i} \right)
# \right\} _{i=0}^{2N}$ that incorporates this new labelling, such that for each dataset
# (indicated by the subscript $D = 0$ or $D=T$),
#
# $$ X_{D,i} = \left( \mathbf{x}_{D,i}, z \right), $$
#
# and
#
# $$ Y_{D,i} = y_{D,i}^{(z)}, $$
#
# where $z = 0$ if $i$ is odd and $z=1$ if $i$ is even.


# %%
# Change vectors x -> X = (x,z), and vectors y -> Y = (y,z) via the artificial z label
def label_position(data):
    # introduce alternating z label
    n_points = len(data[0])
    label = jnp.tile(jnp.array([0.0, 1.0]), n_points)
    return jnp.vstack((jnp.repeat(data, repeats=2, axis=1), label)).T


# change vectors y -> Y by reshaping the velocity measurements
def stack_velocity(data):
    return data.T.flatten().reshape(-1, 1)


def dataset_3d(pos, vel):
    return gpx.Dataset(label_position(pos), stack_velocity(vel))


# label and place the training data into a Dataset object to be used by GPJax
dataset_train = dataset_3d(pos_train, vel_train)

# we also require the testing data to be relabelled for later use, such that we can query the 2Nx2N GP at the test points
dataset_ground_truth = dataset_3d(pos_test, vel_test)


# %% [markdown]
#
# ## Velocity (dimension) decomposition
# Having labelled the data, we are now in a position to use GPJax to learn the function
# $g$, and hence $\mathbf{F}$. A naive approach to the problem is to apply a GP prior
# directly to the velocities of each dimension independently, which is called the
# *velocity* GP. For our prior, we choose an isotropic mean 0 over all dimensions of the
# GP, and a piecewise kernel that depends on the $z$ labels of the inputs, such that for
# two inputs $\mathbf{X} = \left( \mathbf{x}, z \right )$ and $\mathbf{X}^\prime =
# \left( \mathbf{x}^\prime, z^\prime \right )$,
#
# $$ k_{\text{vel}} \left(\mathbf{X}, \mathbf{X}^{\prime}\right)=
# \begin{cases}k^{(z)}\left(\mathbf{x}, \mathbf{x}^{\prime}\right) & \text { if }
# z=z^{\prime} \\ 0 & \text { if } z \neq z^{\prime}, \end{cases} $$
#
# where $k^{(z)}\left(\mathbf{x}, \mathbf{x}^{\prime}\right)$ are the user chosen
# kernels for each dimension. What this means is that there are no correlations between
# the $x^{(0)}$ and $x^{(1)}$ dimensions for all choices $\mathbf{X}$ and
# $\mathbf{X}^{\prime}$, since there are no off-diagonal elements in the Gram matrix
# populated by this choice.
#
# To implement this approach in GPJax, we define `VelocityKernel` in the following cell,
# following the steps outlined in the [custom kernels
# notebook](https://docs.jaxgaussianprocesses.com/_examples/constructing_new_kernels/#custom-kernel).
# This modular implementation takes the choice of user kernels as its class attributes:
# `kernel0` and `kernel1`. We must additionally pass the argument `active_dims = [0,1]`,
# which is an attribute of the base class `AbstractKernel`, into the chosen kernels.
# This is necessary such that the subsequent likelihood optimisation does not optimise
# over the artificial label dimension.
#


# %%


class VelocityKernel(gpx.kernels.AbstractKernel):
    def __init__(
        self,
        kernel0: gpx.kernels.AbstractKernel = gpx.kernels.RBF(active_dims=[0, 1]),
        kernel1: gpx.kernels.AbstractKernel = gpx.kernels.RBF(active_dims=[0, 1]),
    ):
        self.kernel0 = kernel0
        self.kernel1 = kernel1
        super().__init__(compute_engine=DenseKernelComputation())

    def __call__(
        self, X: Float[Array, "1 D"], Xp: Float[Array, "1 D"]
    ) -> Float[Array, "1"]:
        # standard RBF-SE kernel is x and x' are on the same output, otherwise returns 0

        z = jnp.array(X[2], dtype=int)
        zp = jnp.array(Xp[2], dtype=int)

        # achieve the correct value via 'switches' that are either 1 or 0
        k0_switch = ((z + 1) % 2) * ((zp + 1) % 2)
        k1_switch = z * zp

        return k0_switch * self.kernel0(X, Xp) + k1_switch * self.kernel1(X, Xp)


# %% [markdown]
# ### GPJax implementation
# Next, we define the model in GPJax. The prior is defined using
# $k_{\text{vel}}\left(\mathbf{X}, \mathbf{X}^\prime \right)$ and 0 mean and 0
# observation noise. We choose a Gaussian marginal log-likelihood (MLL).
#


# %%
def initialise_gp(kernel, mean, dataset):
    prior = gpx.gps.Prior(mean_function=mean, kernel=kernel)
    likelihood = gpx.likelihoods.Gaussian(
        num_datapoints=dataset.n, obs_stddev=jnp.array([1.0e-3], dtype=jnp.float64)
    )
    posterior = prior * likelihood
    return posterior


# Define the velocity GP
mean = gpx.mean_functions.Zero()
kernel = VelocityKernel()
velocity_posterior = initialise_gp(kernel, mean, dataset_train)


# %% [markdown]
# With a model now defined, we can proceed to optimise the hyperparameters
# of our likelihood over $D_0$. This is done by minimising the MLL using `BFGS`. We also
# plot its value at each step to visually confirm that we have found the minimum. See
# the  [introduction to Gaussian
# Processes](https://docs.jaxgaussianprocesses.com/_examples/intro_to_gps/) notebook for
# more information on optimising the MLL.


# %%
def optimise_mll(posterior, dataset, NIters=1000, key=key):
    # define the MLL using dataset_train
    objective = lambda p, d: -gpx.objectives.conjugate_mll(p, d)
    # Optimise to minimise the MLL
    opt_posterior, history = gpx.fit_scipy(
        model=posterior,
        objective=objective,
        train_data=dataset,
    )
    return opt_posterior


opt_velocity_posterior = optimise_mll(velocity_posterior, dataset_train)


# %% [markdown]
# ### Comparison
# We next obtain the latent distribution of the GP of $g$ at $\mathbf{x}_{0,i}$, then
# extract its mean and standard at the test locations,
# $\mathbf{F}_{\text{latent}}(\mathbf{x}_{0,i})$, as well as the standard deviation (we
# will use it at the very end).


# %%
def latent_distribution(opt_posterior, pos_3d, dataset_train):
    latent = opt_posterior.predict(pos_3d, train_data=dataset_train)
    latent_mean = latent.mean
    latent_std = latent.stddev()
    return latent_mean, latent_std


# extract latent mean and std of g, redistribute into vectors to model F
velocity_mean, velocity_std = latent_distribution(
    opt_velocity_posterior, dataset_ground_truth.X, dataset_train
)

dataset_latent_velocity = dataset_3d(pos_test, velocity_mean)


# %% [markdown]
# We now replot the ground truth (testing data) $D_0$, the predicted
# latent vector field $\mathbf{F}_{\text{latent}}(\mathbf{x_i})$, and a heatmap of the
# residuals at each location $\mathbf{R}(\mathbf{x}_{0,i}) = \mathbf{y}_{0,i} -
# \mathbf{F}_{\text{latent}}(\mathbf{x}_{0,i})$, as well as
# $\left|\left|\mathbf{R}(\mathbf{x}_{0,i})\right|\right|$.


# %%
# Residuals between ground truth and estimate


def plot_vector_field(ax, dataset, **kwargs):
    ax.quiver(
        dataset.X[::2][:, 0],
        dataset.X[::2][:, 1],
        dataset.y[::2],
        dataset.y[1::2],
        **kwargs,
    )


def prepare_ax(ax, X, Y, title, **kwargs):
    ax.set(
        xlim=[X.min() - 0.1, X.max() + 0.1],
        ylim=[Y.min() + 0.1, Y.max() + 0.1],
        aspect="equal",
        title=title,
        ylabel="latitude",
        **kwargs,
    )


def residuals(dataset_latent, dataset_ground_truth):
    return jnp.sqrt(
        (dataset_latent.y[::2] - dataset_ground_truth.y[::2]) ** 2
        + (dataset_latent.y[1::2] - dataset_ground_truth.y[1::2]) ** 2
    )


def plot_fields(
    dataset_ground_truth, dataset_trajectory, dataset_latent, shape=shape, scale=10
):
    X = dataset_ground_truth.X[:, 0][::2]
    Y = dataset_ground_truth.X[:, 1][::2]
    # make figure
    fig, ax = plt.subplots(1, 3, figsize=(12.0, 3.0), sharey=True)

    # ground truth
    plot_vector_field(
        ax[0],
        dataset_ground_truth,
        color=colors[0],
        label="Ocean Current",
        angles="xy",
        scale=scale,
    )
    plot_vector_field(
        ax[0],
        dataset_trajectory,
        color=colors[1],
        label="Drifter",
        angles="xy",
        scale=scale,
    )
    prepare_ax(ax[0], X, Y, "Ground Truth", xlabel="Longitude")

    # Latent estimate of vector field F
    plot_vector_field(ax[1], dataset_latent, color=colors[0], angles="xy", scale=scale)
    plot_vector_field(
        ax[1], dataset_trajectory, color=colors[1], angles="xy", scale=scale
    )
    prepare_ax(ax[1], X, Y, "GP Estimate", xlabel="Longitude")

    # residuals
    residuals_vel = jnp.flip(
        residuals(dataset_latent, dataset_ground_truth).reshape(shape), axis=0
    )
    im = ax[2].imshow(
        residuals_vel,
        extent=[X.min(), X.max(), Y.min(), Y.max()],
        cmap="jet",
        vmin=0,
        vmax=1.0,
        interpolation="spline36",
    )
    plot_vector_field(
        ax[2], dataset_trajectory, color=colors[1], angles="xy", scale=scale
    )
    prepare_ax(ax[2], X, Y, "Residuals", xlabel="Longitude")
    fig.colorbar(im, fraction=0.027, pad=0.04, orientation="vertical")

    fig.legend(
        framealpha=0.0,
        ncols=2,
        fontsize="medium",
        bbox_to_anchor=(0.5, -0.03),
        loc="lower center",
    )


plot_fields(dataset_ground_truth, dataset_train, dataset_latent_velocity)


# %% [markdown]
# From the latent estimate we can see the velocity GP struggles to
# reconstruct features of the ground truth. This is because our construction of the
# kernel placed an independent prior on each physical dimension, which cannot be
# assumed. Therefore, we need a different approach that can implicitly incorporate this
# dependence at a fundamental level. To achieve this we will require a *Helmholtz
# Decomposition*.


# %% [markdown]
# ## Helmholtz decomposition
# In 2 dimensions, a twice continuously differentiable and compactly supported vector
# field $\mathbf{F}: \mathbb{R}^2 \rightarrow \mathbb{R}^2$ can be expressed as the sum
# of the gradient of a scalar potential $\Phi: \mathbb{R}^2 \rightarrow \mathbb{R}$,
# called the potential function, and the vorticity operator of another scalar potential
# $\Psi: \mathbb{R}^2 \rightarrow \mathbb{R}$, called the stream function ([Berlinghieri
# et al. (2023)](https://arxiv.org/pdf/2302.10364.pdf)) such that $$
# \mathbf{F}=\operatorname{grad} \Phi+\operatorname{rot} \Psi, $$ where $$
# \operatorname{grad} \Phi:=\left[\begin{array}{l} \partial \Phi / \partial x^{(0)} \\
# \partial \Phi / \partial x^{(1)} \end{array}\right] \text { and } \operatorname{rot}
# \Psi:=\left[\begin{array}{c} \partial \Psi / \partial x^{(1)} \\
# -\partial \Psi / \partial x^{(0)} \end{array}\right].  $$
#
# This is reminiscent of a 3 dimensional [Helmholtz
# decomposition](https://en.wikipedia.org/wiki/Helmholtz_decomposition).
#
# The 2 dimensional decomposition motivates a different approach: placing priors on
# $\Psi$ and $\Phi$, allowing us to make assumptions directly about fundamental
# properties of $\mathbf{F}$. If we choose independent GP priors such that $\Phi \sim
# \mathcal{G P}\left(0, k_{\Phi}\right)$ and $\Psi \sim \mathcal{G P}\left(0,
# k_{\Psi}\right)$, then $\mathbf{F} \sim \mathcal{G P} \left(0, k_\text{Helm}\right)$
# (since acting linear operations on a GPs give GPs).
#
# For $\mathbf{X}, \mathbf{X}^{\prime} \in \mathbb{R}^2 \times \left\{0,1\right\}$ and
# $z, z^\prime \in \{0,1\}$,
#
# $$ \boxed{ k_{\mathrm{Helm}}\left(\mathbf{x}, \mathbf{x}^{\prime}\right)_{z,z^\prime}
# =  \frac{\partial^2 k_{\Phi}\left(\mathbf{x}, \mathbf{x}^{\prime}\right)}{\partial
# x^{(z)} \partial\left(x^{\prime}\right)^{(z^\prime)}}+(-1)^{z+z^\prime}
# \frac{\partial^2 k_{\Psi}\left(\mathbf{x}, \mathbf{x}^{\prime}\right)}{\partial
# x^{(1-z)} \partial\left(x^{\prime}\right)^{(1-z^\prime)}}}.  $$
#
# where $x^{(z)}$ and $(x^\prime)^{(z^\prime)}$ are the $z$ and $z^\prime$ components of
# $\mathbf{X}$ and ${\mathbf{X}}^{\prime}$ respectively.
#
# We compute the second derivatives using `jax.hessian`. In the following
# implementation, for a kernel $k(\mathbf{x}, \mathbf{x}^{\prime})$, this computes the
# Hessian matrix with respect to the components of $\mathbf{x}$
#
# $$ \frac{\partial^2 k\left(\mathbf{x}, \mathbf{x}^{\prime}\right)}{\partial x^{(z)}
# \partial x^{(z^\prime)}}.  $$
#
# Note that we have operated $\dfrac{\partial}{\partial x^{(z)}}$, *not*
# $\dfrac{\partial}{\partial \left(x^\prime \right)^{(z)}}$, as the boxed equation
# suggests. This is not an issue if we choose stationary kernels $k(\mathbf{x},
# \mathbf{x}^{\prime}) = k(\mathbf{x} - \mathbf{x}^{\prime})$ , as the partial
# derivatives with respect to the components have the following exchange symmetry:
#
# $$ \frac{\partial}{\partial x^{(z)}} = - \frac{\partial}{\partial \left( x^\prime
# \right)^{(z)}}, $$
#
# for either $z$.
# %%


@dataclass
class HelmholtzKernel(gpx.kernels.stationary.StationaryKernel):
    # initialise Phi and Psi kernels as any stationary kernel in gpJax
    potential_kernel: gpx.kernels.stationary.StationaryKernel = field(
        default_factory=lambda: gpx.kernels.RBF(active_dims=[0, 1])
    )
    stream_kernel: gpx.kernels.stationary.StationaryKernel = field(
        default_factory=lambda: gpx.kernels.RBF(active_dims=[0, 1])
    )
    compute_engine = DenseKernelComputation()

    def __call__(
        self, X: Float[Array, "1 D"], Xp: Float[Array, "1 D"]
    ) -> Float[Array, "1"]:
        # obtain indices for k_helm, implement in the correct sign between the derivatives
        z = jnp.array(X[2], dtype=int)
        zp = jnp.array(Xp[2], dtype=int)
        sign = (-1) ** (z + zp)

        # convert to array to correctly index, -ve sign due to exchange symmetry (only true for stationary kernels)
        potential_dvtve = -jnp.array(
            hessian(self.potential_kernel)(X, Xp), dtype=jnp.float64
        )[z][zp]
        stream_dvtve = -jnp.array(
            hessian(self.stream_kernel)(X, Xp), dtype=jnp.float64
        )[1 - z][1 - zp]

        return potential_dvtve + sign * stream_dvtve


# %% [markdown]
# ### GPJax implementation
# We repeat the same steps as with the velocity GP model, replacing `VelocityKernel`
# with `HelmholtzKernel`.

# %%
# Redefine Gaussian process with Helmholtz kernel
kernel = HelmholtzKernel()
helmholtz_posterior = initialise_gp(kernel, mean, dataset_train)
# Optimise hyperparameters using BFGS
opt_helmholtz_posterior = optimise_mll(helmholtz_posterior, dataset_train)


# %% [markdown]
# ### Comparison
# We again plot the ground truth (testing data) $D_0$, the predicted latent vector field
# $\mathbf{F}_{\text{latent}}(\mathbf{x}_{0,i})$, and a heatmap of the residuals at each
# location $R(\mathbf{x}_{0,i}) = \mathbf{y}_{0,i} -
# \mathbf{F}_{\text{latent}}(\mathbf{x}_{0,i})$ and $\left|\left|R(\mathbf{x}_{0,i})
# \right|\right|$.

# %%
# obtain latent distribution, extract x and y values over g
helmholtz_mean, helmholtz_std = latent_distribution(
    opt_helmholtz_posterior, dataset_ground_truth.X, dataset_train
)
dataset_latent_helmholtz = dataset_3d(pos_test, helmholtz_mean)

plot_fields(dataset_ground_truth, dataset_train, dataset_latent_helmholtz)

# %% [markdown]
# Visually, the Helmholtz model performs better than the velocity model, preserving the
# local structure of the $\mathbf{F}$. Since we placed priors on $\Phi$ and $\Psi$, the
# construction of $\mathbf{F}$ allows for correlations between the dimensions (non-zero
# off-diagonal elements in the Gram matrix populated by
# $k_\text{Helm}\left(\mathbf{X},\mathbf{X}^{\prime}\right)$ ).


# %% [markdown]
# ## Negative log predictive densities
# Lastly, we directly compare the velocity and Helmholtz models by computing the
# [negative log predictive
# densities](https://en.wikipedia.org/wiki/Negative_log_predictive_density) for each
# model. This is a quantitative metric that measures the probability of the ground truth
# given the data.
#
# $$
# \mathrm{NLPD}=-\sum_{i=1}^{2N} \log \left(  p\left(\mathcal{Y}_i = Y_{0,i} \mid \mathbf{X}_{i}\right) \right),
# $$
#
# where each $p\left(\mathcal{Y}_i \mid \mathbf{X}_i \right)$ is the marginal Gaussian
# distribution over $\mathcal{Y}_i$ at each test location, and $Y_{i,0}$ is the $i$-th
# component of the (massaged) test data that we reserved at the beginning of the
# notebook in $D_0$. A smaller value is better, since the deviation of the ground truth
# and the model are small in this case.


# %%
# ensure testing data alternates between x0 and x1 components
def nlpd(mean, std, vel_test):
    vel_query = jnp.column_stack((vel_test[0], vel_test[1])).flatten()
    normal = npd.Normal(loc=mean, scale=std)
    return -jnp.sum(normal.log_prob(vel_query))


# compute nlpd for velocity and helmholtz
nlpd_vel = nlpd(velocity_mean, velocity_std, vel_test)
nlpd_helm = nlpd(helmholtz_mean, helmholtz_std, vel_test)

print("NLPD for Velocity: %.2f \nNLPD for Helmholtz: %.2f" % (nlpd_vel, nlpd_helm))
# %% [markdown]
# The Helmholtz model outperforms the velocity model, as indicated by the lower NLPD score.

# %% [markdown]
# <span id="fn1"></span>
# ## Footnote
# Kernels for vector-valued functions have been studied in the literature, see [Alvarez et al. (2012)](https://doi.org/10.48550/arXiv.1106.6251)
# ## System configuration
# %%
# %reload_ext watermark
# %watermark -n -u -v -iv -w -a 'Ivan Shalashilin'



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import beartype.typing as tp
from jax import vmap
from jaxtyping import Float

import gpjax  # noqa: F401
from gpjax.kernels.computations.base import AbstractKernelComputation
from gpjax.typing import Array

K = tp.TypeVar("K", bound="gpjax.kernels.base.AbstractKernel")  # noqa: F821


class DenseKernelComputation(AbstractKernelComputation):
    r"""Dense kernel computation class. Operations with the kernel assume
    a dense gram matrix structure.
    """

    def _cross_covariance(
        self, kernel: K, x: Float[Array, "N D"], y: Float[Array, "M D"]
    ) -> Float[Array, "N M"]:
        cross_cov = vmap(lambda x: vmap(lambda y: kernel(x, y))(y))(x)
        return cross_cov



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import jax.numpy as jnp
from jaxtyping import Float
import numpyro.distributions as npd

from gpjax.kernels.stationary.base import StationaryKernel
from gpjax.kernels.stationary.utils import (
    build_student_t_distribution,
    euclidean_distance,
)
from gpjax.typing import (
    Array,
    ScalarFloat,
)


class Matern12(StationaryKernel):
    r"""The Matérn kernel with smoothness parameter fixed at 0.5.

    Computes the covariance on a pair of inputs $(x, y)$ with
    lengthscale parameter $\ell$ and variance $\sigma^2$.

    $$
    k(x, y) = \sigma^2\exp\Bigg(-\frac{\lvert x-y \rvert}{2\ell^2}\Bigg)
    $$
    """

    name: str = "Matérn12"

    def __call__(self, x: Float[Array, " D"], y: Float[Array, " D"]) -> ScalarFloat:
        x = self.slice_input(x) / self.lengthscale.value
        y = self.slice_input(y) / self.lengthscale.value
        K = self.variance.value * jnp.exp(-euclidean_distance(x, y))
        return K.squeeze()

    @property
    def spectral_density(self) -> npd.StudentT:
        return build_student_t_distribution(nu=1)



from typing import Tuple

from cola.ops import Dense
import jax
from jax import config
import jax.numpy as jnp
import jax.random as jr
import pytest

from gpjax.kernels.approximations import RFF
from gpjax.kernels.nonstationary import (
    Linear,
    Polynomial,
)
from gpjax.kernels.stationary import (
    RBF,
    Matern12,
    Matern32,
    Matern52,
    Periodic,
    PoweredExponential,
    RationalQuadratic,
    StationaryKernel,
)

config.update("jax_enable_x64", True)
_jitter = 1e-6


@pytest.mark.parametrize("kernel", [RBF, Matern12, Matern32, Matern52])
@pytest.mark.parametrize("num_basis_fns", [2, 10])
@pytest.mark.parametrize("n_dims", [1, 3])
def test_frequency_sampler(
    kernel: type[StationaryKernel], num_basis_fns: int, n_dims: int
):
    base_kernel = kernel(active_dims=list(range(n_dims)))
    approximate = RFF(base_kernel=base_kernel, num_basis_fns=num_basis_fns)
    assert approximate.frequencies.value.shape == (num_basis_fns, n_dims)


@pytest.mark.parametrize("kernel", [RBF, Matern12, Matern32, Matern52])
@pytest.mark.parametrize("num_basis_fns", [2, 10])
@pytest.mark.parametrize("n_dims", [1, 3])
@pytest.mark.parametrize("n_data", [50, 100])
def test_gram(
    kernel: type[StationaryKernel], num_basis_fns: int, n_dims: int, n_data: int
):
    key = jr.PRNGKey(123)
    x = jr.uniform(key, shape=(n_data, 1), minval=-3.0, maxval=3.0).reshape(-1, 1)
    if n_dims > 1:
        x = jnp.hstack([x] * n_dims)
    base_kernel = kernel(active_dims=list(range(n_dims)))
    approximate = RFF(base_kernel=base_kernel, num_basis_fns=num_basis_fns)

    linop = approximate.gram(x)

    # Check the return type
    assert isinstance(linop, Dense)

    Kxx = linop.to_dense() + jnp.eye(n_data) * _jitter

    # Check that the shape is correct
    assert Kxx.shape == (n_data, n_data)

    # Check that the Gram matrix is PSD
    evals, _ = jnp.linalg.eigh(Kxx)
    assert jnp.all(evals > 0)


@pytest.mark.parametrize("kernel", [RBF, Matern12, Matern32, Matern52])
@pytest.mark.parametrize("num_basis_fns", [2, 10])
@pytest.mark.parametrize("n_dims", [1, 3])
@pytest.mark.parametrize("n_datas", [(50, 100), (100, 50)])
def test_cross_covariance(
    kernel: type[StationaryKernel],
    num_basis_fns: int,
    n_dims: int,
    n_datas: Tuple[int, int],
):
    nd1, nd2 = n_datas
    key = jr.key(123)
    x1 = jr.uniform(key, shape=(nd1, 1), minval=-3.0, maxval=3.0)
    if n_dims > 1:
        x1 = jnp.hstack([x1] * n_dims)
    x2 = jr.uniform(key, shape=(nd2, 1), minval=-3.0, maxval=3.0)
    if n_dims > 1:
        x2 = jnp.hstack([x2] * n_dims)

    base_kernel = kernel(active_dims=list(range(n_dims)))
    approximate = RFF(base_kernel=base_kernel, num_basis_fns=num_basis_fns)
    Kxx = approximate.cross_covariance(x1, x2)

    # Check the return type
    assert isinstance(Kxx, jax.Array)

    # Check that the shape is correct
    assert Kxx.shape == (nd1, nd2)


@pytest.mark.parametrize("kernel", [RBF, Matern12, Matern32, Matern52])
@pytest.mark.parametrize("n_dim", [1, 3])
def test_improvement(kernel: type[StationaryKernel], n_dim: int):
    n_data = 100
    key = jr.key(123)

    x = jr.uniform(key, minval=-3.0, maxval=3.0, shape=(n_data, n_dim))
    base_kernel = kernel(active_dims=list(range(n_dim)))
    exact_linop = base_kernel.gram(x).to_dense()

    crude_approximation = RFF(base_kernel=base_kernel, num_basis_fns=10)
    c_linop = crude_approximation.gram(x).to_dense()

    better_approximation = RFF(base_kernel=base_kernel, num_basis_fns=100)
    b_linop = better_approximation.gram(x).to_dense()

    c_delta = jnp.linalg.norm(exact_linop - c_linop, ord="fro")
    b_delta = jnp.linalg.norm(exact_linop - b_linop, ord="fro")

    # The frobenius norm of the difference between the exact and approximate
    # should improve as we increase the number of basis functions
    assert c_delta > b_delta


@pytest.mark.parametrize("kernel", [RBF, Matern12, Matern32, Matern52])
def test_exactness(kernel: type[StationaryKernel]):
    kernel = kernel(n_dims=1)

    n_data = 100
    key = jr.key(123)

    x = jr.uniform(key, minval=-3.0, maxval=3.0, shape=(n_data, 1))
    exact_linop = kernel.gram(x).to_dense()

    better_approximation = RFF(base_kernel=kernel, num_basis_fns=300)
    b_linop = better_approximation.gram(x).to_dense()

    max_delta = jnp.max(exact_linop - b_linop)
    assert max_delta < 0.1


@pytest.mark.parametrize("kernel", [Polynomial, Linear])
def test_nonstationary_raises_error(kernel):
    with pytest.raises(TypeError):
        RFF(base_kernel=kernel(1), num_basis_fns=10)


@pytest.mark.parametrize(
    "kernel",
    [RationalQuadratic, PoweredExponential, Periodic],
)
def test_missing_spectral_density_raises_error(kernel):
    with pytest.raises(NotImplementedError):
        RFF(base_kernel=kernel(), num_basis_fns=10)


@pytest.mark.parametrize("kernel", [RBF, Matern12, Matern32, Matern52])
def test_stochastic_init(kernel: type[StationaryKernel]):
    with pytest.raises(ValueError):
        # n_dims is not specified, but should be
        RFF(base_kernel=kernel(), num_basis_fns=10)

    kernel = kernel(n_dims=1)

    k1 = RFF(base_kernel=kernel, num_basis_fns=10, key=jr.PRNGKey(123))
    k2 = RFF(base_kernel=kernel, num_basis_fns=10, key=jr.PRNGKey(42))

    assert (k1.frequencies.value != k2.frequencies.value).any()



from jax import config
from jaxtyping import install_import_hook

config.update("jax_enable_x64", True)

# import gpjax within import hook to apply beartype everywhere, before running tests
with install_import_hook("gpjax", "beartype.beartype"):
    import gpjax  # noqa: F401



# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     custom_cell_magics: kql
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.7
#   kernelspec:
#     display_name: gpjax
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Introduction to Kernels

# %% [markdown]
# In this guide we provide an introduction to kernels, and the role they play in Gaussian process models.

# %%
# Enable Float64 for more stable matrix inversions.
from jax import config
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import (
    Float,
    install_import_hook,
)
import matplotlib as mpl
import matplotlib.pyplot as plt
import optax as ox
import pandas as pd
from sklearn.preprocessing import StandardScaler

from examples.utils import use_mpl_style
from gpjax.parameters import Static
from gpjax.typing import Array

config.update("jax_enable_x64", True)


with install_import_hook("gpjax", "beartype.beartype"):
    import gpjax as gpx


key = jr.key(42)

# set the default style for plotting
use_mpl_style()
cols = mpl.rcParams["axes.prop_cycle"].by_key()["color"]

# %% [markdown]
# Using Gaussian Processes (GPs) to model functions can offer several advantages over
# alternative methods, such as deep neural networks. One key advantage is their rich
# quantification of uncertainty; not only do they provide *point estimates* for the
# values taken by a function throughout its domain, but they provide a full predictive
# posterior *distribution* over the range of values the function may take. This rich
# quantification of uncertainty is useful in many applications, such as Bayesian
# optimisation, which relies on being able to make *uncertainty-aware* decisions.
#
# However, another advantage of GPs is the ability for one to place *priors* on the
# functions being modelled. For instance, one may know that the underlying function
# being modelled observes certain characteristics, such as being *periodic* or having a
# certain level of *smoothness*. The *kernel*, or *covariance function*, is the primary
# means through which one is able to encode such prior knowledge about the function
# being modelled. This enables one to equip the GP with inductive biases which enable it
# to learn from data more efficiently, whilst generalising to unseen data more
# effectively.
#
# In this notebook we'll develop some intuition for what kinds of priors are encoded
# through the use of different kernels, and how this can be useful when modelling
# different types of functions.

# %% [markdown]
# ## What is a Kernel?
#
# Intuitively, for a function $f$, the kernel defines the notion of *similarity* between
# the value of the function at two points, $f(\mathbf{x})$ and $f(\mathbf{x}')$, and
# will be denoted as $k(\mathbf{x}, \mathbf{x}')$:
#
# $$
# \begin{aligned}
#   k(\mathbf{x}, \mathbf{x}') &= \text{Cov}[f(\mathbf{x}), f(\mathbf{x}')] \\
#   &= \mathbb{E}[(f(\mathbf{x}) - \mathbb{E}[f(\mathbf{x})])(f(\mathbf{x}') - \mathbb{E}[f(\mathbf{x}')])]
# \end{aligned}
# $$
#
# One would expect that, given a previously unobserved test point $\mathbf{x}^*$, the
# training points which are *closest* to this unobserved point will be most similar to
# it. As such, the kernel is used to define this notion of similarity within the GP
# framework. It is up to the user to select a kernel function which is appropriate for
# the function being modelled. In this notebook we are going to give some examples of
# commonly used kernels, and try to develop an understanding of when one may wish to use
# one kernel over another. However, before we do this, it is worth discussing the
# necessary conditions for a function to be a valid kernel/covariance function. This
# requires a little bit of maths, so for those of you who just wish to obtain an
# intuitive understanding, feel free to skip to the section introducing the Matérn
#  family of kernels.
#
# ### What are the necessary conditions for a function to be a valid kernel?
#
# Whilst intuitively the kernel function is used to define the notion of similarity within
# the GP framework, it is important to note that there are two *necessary conditions*
# that a kernel function must satisfy in order to be a valid covariance function. For
# clarity, we will refer to *any* function mapping two inputs to a scalar output as a
# *kernel function*, and we will refer to a *valid* kernel function satisfying the two
# necessary conditions as a *covariance function*. However, it is worth noting that the
# GP community often uses the terms *kernel function* and *covariance function*
# interchangeably.
#
# The first necessary condition is that the covariance function must be *symmetric*, i.e.
# $k(\mathbf{x}, \mathbf{x}') = k(\mathbf{x}', \mathbf{x})$. This is because the
# covariance between two random variables $X$ and $X'$ is symmetric; if one looks at the
# definition of covariance given above, it is clear that it is invariant to swapping the
# order of the inputs $\mathbf{x}$ and $\mathbf{x}'$.
#
# The second necessary condition is that the covariance function must be *positive
# semi-definite* (PSD). In order to understand this condition, it is useful to first
# introduce the concept of a *Gram matrix*. We'll use the same notation as the [GP introduction
# notebook](https://docs.jaxgaussianprocesses.com/_examples/intro_to_gps/), and denote
# $n$ input points as $\mathbf{X} = \{\mathbf{x}_1, \ldots, \mathbf{x}_n\}$. Given these
# input points and a kernel function $k$ the *Gram matrix* stores the pairwise kernel
# evaluations between all input points. Mathematically, this leads to the Gram matrix being defined as:
#
# $$K(\mathbf{X}, \mathbf{X}) = \begin{bmatrix} k(\mathbf{x}_1, \mathbf{x}_1) & \cdots & k(\mathbf{x}_1, \mathbf{x}_n) \\ \vdots & \ddots & \vdots \\ k(\mathbf{x}_n, \mathbf{x}_1) & \cdots & k(\mathbf{x}_n, \mathbf{x}_n) \end{bmatrix}$$
#
# such that $K(\mathbf{X}, \mathbf{X})_{ij} = k(\mathbf{x}_i, \mathbf{x}_j)$.
#
# In order for $k$ to be a valid covariance function, the corresponding Gram matrix
# must be *positive semi-definite*. In this case the Gram matrix is referred to as a
# *covariance matrix*. A real $n \times n$ matrix $K$ is positive semi-definite if and
# only if for all vectors $\mathbf{z} \in \mathbb{R}^n$:
#
# $$\mathbf{z}^\top K \mathbf{z} \geq 0$$
#
# Alternatively, a real $n \times n$ matrix $K$ is positive semi-definite if and only if
# all of its eigenvalues are non-negative.
#
# Therefore, the two necessary conditions for a function to be a valid covariance function
# are that it must be *symmetric* and *positive semi-definite*. In this section we have
# referred to *any* function from two inputs to a scalar output as a *kernel function*,
# with its corresponding matrix of pairwise evaluations referred to as the *Gram matrix*,
# and a function satisfying the two necessary conditions as a *covariance function*, with
# its corresponding matrix of pairwise evaluations referred to as the *covariance matrix*.
# This enabled us to easily define the necessary conditions for a function to be a valid
# covariance function. However, as noted previously, the GP community often uses these
# terms interchangeably, and so we will for the remainder of this notebook.
#

# %% [markdown]
# ## Introducing a Common Family of Kernels - The Matérn Family

# %% [markdown]
# One of the most widely used families of kernels is the Matérn family ([Matérn, 1960](https://core.ac.uk/download/pdf/11698705.pdf)). These kernels take on the following form:
#
# $$k_{\nu}(\mathbf{x}, \mathbf{x'}) = \sigma^2 \frac{2^{1 - \nu}}{\Gamma(\nu)}\left(\sqrt{2\nu} \frac{|\mathbf{x} - \mathbf{x'}|}{\kappa}\right)^{\nu} K_{\nu} \left(\sqrt{2\nu} \frac{|\mathbf{x} - \mathbf{x'}|}{\kappa}\right)$$
#
# where $K_{\nu}$ is a modified Bessel function, $\nu$, $\kappa$ and $\sigma^2$ are
# hyperparameters specifying the mean-square differentiability, lengthscale and
# variability respectively, and $|\cdot|$ is used to denote the Euclidean norm. Note that
# for those of you less interested in the mathematical underpinnings of kernels, it isn't
# necessary to understand the exact functional form of the Matérn kernels to
# gain an understanding of how they behave. The key takeaway is that they are
# parameterised by several hyperparameters, and that these hyperparameters dictate the
# behaviour of functions sampled from the corresponding GP. The plots below will provide
# some more intuition for how these hyperparameters affect the behaviour of functions
# sampled from the corresponding GP.
#
#
# Some commonly used Matérn kernels use half-integer values of $\nu$, such as $\nu = 1/2$
# or $\nu = 3/2$. The fraction is sometimes omitted when naming the kernel, so that $\nu =
# 1/2$ is referred to as the Matérn12 kernel, and $\nu = 3/2$ is referred to as the
# Matérn32 kernel. When $\nu$ takes in a half-integer value, $\nu = k + 1/2$, the kernel
# can be expressed as the product of a polynomial of order $k$ and an exponential:
#
# $$k_{k + 1/2}(\mathbf{x}, \mathbf{x'}) = \sigma^2
# \exp\left(-\frac{\sqrt{2\nu}|\mathbf{x} - \mathbf{x'}|}{\kappa}\right)
# \frac{\Gamma(k+1)}{\Gamma(2k+1)} \times \sum_{i= 0}^k \frac{(k+i)!}{i!(k-i)!}
# \left(\frac{(\sqrt{8\nu}|\mathbf{x} - \mathbf{x'}|)}{\kappa}\right)^{k-i}$$
#
# In the limit of $\nu \to \infty$ this yields the *squared-exponential*, or *radial basis function (RBF)*, kernel, which is infinitely mean-square differentiable:
#
# $$k_{\infty}(\mathbf{x}, \mathbf{x'}) = \sigma^2 \exp\left(-\frac{|\mathbf{x} - \mathbf{x'}|^2}{2\kappa^2}\right)$$
#
# But what kind of functions does this kernel encode prior knowledge about? Let's take a look at some samples from GP priors defined used Matérn kernels with different values of $\nu$:

# %%
kernels = [
    gpx.kernels.Matern12(),
    gpx.kernels.Matern32(),
    gpx.kernels.Matern52(),
    gpx.kernels.RBF(),
]
fig, axes = plt.subplots(ncols=2, nrows=2, figsize=(7, 6), tight_layout=True)

x = jnp.linspace(-3.0, 3.0, num=200).reshape(-1, 1)

meanf = gpx.mean_functions.Zero()

for k, ax in zip(kernels, axes.ravel(), strict=False):
    prior = gpx.gps.Prior(mean_function=meanf, kernel=k)
    rv = prior(x)
    y = rv.sample(key=key, sample_shape=(10,))
    ax.plot(x, y.T, alpha=0.7)
    ax.set_title(k.name)


# %% [markdown]
# The plots above clearly show that the choice of $\nu$ has a large impact on the *smoothness* of the functions being modelled by the GP, with functions drawn from GPs defined with the Matérn kernel becoming increasingly smooth as $\nu \to \infty$. More formally, this notion of smoothness is captured through the mean-square differentiability of the function being modelled. Functions sampled from GPs using a Matérn kernel are $k$-times mean-square differentiable, if and only if $\nu > k$. For instance, functions sampled from a GP using a Matérn12 kernel are zero times mean-square differentiable, and functions sampled from a GP using the RBF kernel are infinitely mean-square differentiable.
#
# As an important aside, a general property of the Matérn family of kernels is that they are examples of *stationary* kernels. This means that they only depend on the *displacement* of the two points being compared, $\mathbf{x} - \mathbf{x}'$, and not on their absolute values. This is a useful property to have, as it means that the kernel is invariant to translations in the input space. They also go beyond this, as they only depend on the Euclidean *distance* between the two points being compared, $|\mathbf{x} - \mathbf{x}'|$. Kernels which satisfy this property are known as *isotropic* kernels. This makes the function invariant to all rigid motions in the input space, such as rotations.

# %% [markdown]
# ## Inferring Kernel Hyperparameters

# %% [markdown]
# Most kernels have several *hyperparameters*, which we denote $\mathbf{\theta}$, which encode different assumptions about the underlying function being modelled. For the Matérn family described above, $\mathbf{\theta} = \{\nu, \kappa, \sigma\}$. A fully Bayesian approach to dealing with hyperparameters would be to place a prior over them, and marginalise over the posterior derived from the data in order to perform predictions. However, this is often computationally very expensive, and so a common approach is to instead *optimise* the hyperparameters by maximising the log marginal likelihood of the data. Given training data $\mathbf{D} = (\mathbf{X}, \mathbf{y})$, assumed to contain some additive Gaussian noise $\epsilon \sim \mathcal{N}(0, \sigma^2)$, the log marginal likelihood of the dataset is defined as:
#
# $$ \begin{aligned}
# \log(p(\mathbf{y} | \mathbf{X}, \boldsymbol{\theta})) &= \log\left(\int p(\mathbf{y} | \mathbf{f}, \mathbf{X}, \boldsymbol{\theta}) p(\mathbf{f} | \mathbf{X}, \boldsymbol{\theta}) d\mathbf{f}\right) \nonumber \\
# &= - \frac{1}{2} \mathbf{y} ^ \top \left(K(\mathbf{X}, \mathbf{X}) + \sigma^2 \mathbf{I} \right)^{-1} \mathbf{y} - \frac{1}{2} \log |K(\mathbf{X}, \mathbf{X}) + \sigma^2 \mathbf{I}| - \frac{n}{2} \log 2 \pi
# \end{aligned}$$

# %% [markdown]
# This expression can then be maximised with respect to the hyperparameters using a
# gradient-based approach such as Adam or L-BFGS. Note that we may choose to fix some
# hyperparameters, and in GPJax the parameter $\nu$ is set by the user, and not
# inferred though optimisation. For more details on using the log marginal likelihood to
# optimise kernel hyperparameters, see our [GP introduction notebook](https://docs.jaxgaussianprocesses.com/_examples/intro_to_gps/#gaussian-process-regression).
#
# We'll demonstrate the advantages of being able to infer kernel parameters from the training data by fitting a GP to the widely used [Forrester function](https://www.sfu.ca/~ssurjano/forretal08.html):
#
# $$f(x) = (6x - 2)^2 \sin(12x - 4)$$


# %%
# Forrester function
def forrester(x: Float[Array, "N"]) -> Float[Array, "N"]:  # noqa: F821
    return (6 * x - 2) ** 2 * jnp.sin(12 * x - 4)


n = 13

training_x = jr.uniform(key=key, minval=0, maxval=1, shape=(n,)).reshape(-1, 1)
training_y = forrester(training_x)
D = gpx.Dataset(X=training_x, y=training_y)

test_x = jnp.linspace(0, 1, 100).reshape(-1, 1)
test_y = forrester(test_x)

# %% [markdown]
# First we define our model, using the Matérn52 kernel, and construct our posterior *without* optimising the kernel hyperparameters:

# %%
mean = gpx.mean_functions.Zero()
kernel = gpx.kernels.Matern52(
    lengthscale=jnp.array(0.1)
)  # Initialise our kernel lengthscale to 0.1

prior = gpx.gps.Prior(mean_function=mean, kernel=kernel)

likelihood = gpx.likelihoods.Gaussian(
    num_datapoints=D.n, obs_stddev=Static(jnp.array(1e-3))
)  # Our function is noise-free, so we set the observation noise's standard deviation to a very small value

no_opt_posterior = prior * likelihood

# %% [markdown]
# We can then optimise the hyperparameters by minimising the negative log marginal likelihood of the data:

# %%
gpx.objectives.conjugate_mll(no_opt_posterior, data=D)


# %%
opt_posterior, history = gpx.fit_scipy(
    model=no_opt_posterior,
    objective=lambda p, d: -gpx.objectives.conjugate_mll(p, d),
    train_data=D,
)


# %% [markdown]
# Having optimised the hyperparameters, we can now make predictions using the posterior
# with the optimised hyperparameters, and compare them to the predictions made using the
# posterior with the default hyperparameters:


# %%
def plot_ribbon(ax, x, dist, color):
    mean = dist.mean
    std = jnp.sqrt(dist.variance)
    ax.plot(x, mean, label="Predictive mean", color=color)
    ax.fill_between(
        x.squeeze(),
        mean - 2 * std,
        mean + 2 * std,
        alpha=0.2,
        label="Two sigma",
        color=color,
    )
    ax.plot(x, mean - 2 * std, linestyle="--", linewidth=1, color=color)
    ax.plot(x, mean + 2 * std, linestyle="--", linewidth=1, color=color)


# %%
opt_latent_dist = opt_posterior.predict(test_x, train_data=D)
opt_predictive_dist = opt_posterior.likelihood(opt_latent_dist)

opt_predictive_mean = opt_predictive_dist.mean
opt_predictive_std = jnp.sqrt(opt_predictive_dist.variance)

fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(5, 6))
ax1.plot(
    test_x, test_y, label="Latent function", color=cols[0], linestyle="--", linewidth=2
)
ax1.plot(training_x, training_y, "x", label="Observations", color="k", zorder=5)
plot_ribbon(ax1, test_x, opt_predictive_dist, color=cols[1])
ax1.set_title("Posterior with Hyperparameter Optimisation")
ax1.legend(loc="center left", bbox_to_anchor=(0.975, 0.5))

no_opt_latent_dist = no_opt_posterior.predict(test_x, train_data=D)
no_opt_predictive_dist = no_opt_posterior.likelihood(no_opt_latent_dist)

ax2.plot(
    test_x, test_y, label="Latent function", color=cols[0], linestyle="--", linewidth=2
)
ax2.plot(training_x, training_y, "x", label="Observations", color="k", zorder=5)
plot_ribbon(ax2, test_x, no_opt_predictive_dist, color=cols[1])
ax2.set_title("Posterior without Hyperparameter Optimisation")
ax2.legend(loc="center left", bbox_to_anchor=(0.975, 0.5))

# %% [markdown]
# We can see that optimising the hyperparameters by minimising the negative log marginal likelihood of the data results in a more faithful fit of the GP to the data. In particular, we can observe that the GP using optimised hyperparameters is more accurately able to reflect uncertainty in its predictions, as opposed to the GP using the default parameters, which is overconfident in its predictions.
#
# The lengthscale, $\kappa$, and variance, $\sigma^2$, are shown below, both before and after optimisation:

# %%
no_opt_lengthscale = no_opt_posterior.prior.kernel.lengthscale
no_opt_variance = no_opt_posterior.prior.kernel.variance
opt_lengthscale = opt_posterior.prior.kernel.lengthscale
opt_variance = opt_posterior.prior.kernel.variance

print(f"Optimised Lengthscale: {opt_lengthscale} and Variance: {opt_variance}")
print(
    f"Non-Optimised Lengthscale: {no_opt_lengthscale} and Variance: {no_opt_variance}"
)

# %% [markdown]
# ## Expressing Other Priors with Different Kernels

# %% [markdown]
# Whilst the Matérn kernels are often used as a first choice of kernel, and they often perform well due to their smoothing properties often being well-aligned with the properties of the underlying function being modelled, sometimes more prior knowledge is known about the function being modelled. For instance, it may be known that the function being modelled is *periodic*. In this case, a suitable kernel choice would be the *periodic* kernel:
#
# $$k(\mathbf{x}, \mathbf{x}') = \sigma^2 \exp \left( -\frac{1}{2} \sum_{i=1}^{D} \left(\frac{\sin (\pi (\mathbf{x}_i - \mathbf{x}_i')/p)}{\ell}\right)^2 \right)$$
#
# with $D$ being the dimensionality of the inputs.
#
# Below we show $10$ samples drawn from a GP prior using the periodic kernel:

# %%
mean = gpx.mean_functions.Zero()
kernel = gpx.kernels.Periodic()
prior = gpx.gps.Prior(mean_function=mean, kernel=kernel)

x = jnp.linspace(-3.0, 3.0, num=200).reshape(-1, 1)
rv = prior(x)
y = rv.sample(key=key, sample_shape=(10,))

fig, ax = plt.subplots()
ax.plot(x, y.T, alpha=0.7)
ax.set_title("Samples from the Periodic Kernel")

# %% [markdown]
# In other scenarios, it may be known that the underlying function is *linear*, in which case the *linear* kernel would be a suitable choice:
#
# $$k(\mathbf{x}, \mathbf{x}') = \sigma^2 \mathbf{x}^\top \mathbf{x}'$$
#
# Unlike the kernels shown above, the linear kernel is *not* stationary, and so it is not invariant to translations in the input space.
#
# Below we show $10$ samples drawn from a GP prior using the linear kernel:

# %%
mean = gpx.mean_functions.Zero()
kernel = gpx.kernels.Linear()
prior = gpx.gps.Prior(mean_function=mean, kernel=kernel)

x = jnp.linspace(-3.0, 3.0, num=200).reshape(-1, 1)
rv = prior(x)
y = rv.sample(key=key, sample_shape=(10,))

fig, ax = plt.subplots()
ax.plot(x, y.T, alpha=0.7)
ax.set_title("Samples from the Linear Kernel")

# %% [markdown]
# ## Composing Kernels

# %% [markdown]
# It is also mathematically valid to compose kernels through operations such as addition
# and multiplication in order to produce more expressive kernels. For the mathematically
# interested amongst you, this is valid as the resulting kernel functions still satisfy
# the necessary conditions introduced at the [start of this
# notebook](#what-are-the-necessary-conditions-for-a-function-to-be-a-valid-kernel).
# Adding or multiplying kernel functions is equivalent to performing elementwise addition
# or multiplication of the corresponding covariance matrices, and fortunately symmetric,
# positive semi-definite kernels are closed under these operations. This means that
# kernels produced by adding or multiplying other kernels will also be symmetric and
# positive semi-definite, and so will also be valid kernels. GPJax provides the
# functionality required to easily compose kernels via addition and multiplication, which
# we'll demonstrate below.
#
# First, we'll take a look at some samples drawn from a GP prior using a kernel which is
# composed of the sum of a linear kernel and a periodic kernel:

# %%
kernel_one = gpx.kernels.Linear()
kernel_two = gpx.kernels.Periodic()
sum_kernel = gpx.kernels.SumKernel(kernels=[kernel_one, kernel_two])
mean = gpx.mean_functions.Zero()
prior = gpx.gps.Prior(mean_function=mean, kernel=sum_kernel)

x = jnp.linspace(-3.0, 3.0, num=200).reshape(-1, 1)
rv = prior(x)
y = rv.sample(key=key, sample_shape=(10,))
fig, ax = plt.subplots()
ax.plot(x, y.T, alpha=0.7)
ax.set_title("Samples from a GP Prior with Kernel = Linear + Periodic")


# %% [markdown]
# We can see that the samples drawn behave as one would naturally expect through adding
# the two kernels together. In particular, the samples are still periodic, as with the
# periodic kernel, but their mean also linearly increases/decreases as they move away from
# the origin, as seen with the linear kernel.
#
# Below we take a look at some samples drawn from a GP prior using a kernel which is
# composed of the same two kernels, but this time multiplied together:

# %%
kernel_one = gpx.kernels.Linear()
kernel_two = gpx.kernels.Periodic()
sum_kernel = gpx.kernels.ProductKernel(kernels=[kernel_one, kernel_two])
mean = gpx.mean_functions.Zero()
prior = gpx.gps.Prior(mean_function=mean, kernel=sum_kernel)

x = jnp.linspace(-3.0, 3.0, num=200).reshape(-1, 1)
rv = prior(x)
y = rv.sample(key=key, sample_shape=(10,))
fig, ax = plt.subplots()
ax.plot(x, y.T, alpha=0.7)
ax.set_title("Samples from a GP with Kernel = Linear x Periodic")


# %% [markdown]
# Once again, the samples drawn behave as one would naturally expect through multiplying
# the two kernels together. In particular, the samples are still periodic but their mean
# linearly increases/decreases as they move away from the origin, and the amplitude of
# the oscillations also linearly increases with increasing distance from the origin.

# %% [markdown]
# ## Putting it All Together on a Real-World Dataset

# %% [markdown]
# ### Mauna Loa CO2 Dataset

# %% [markdown]
# We'll put together some of the ideas we've discussed in this notebook by fitting a GP
# to the [Mauna Loa CO2 dataset](https://www.esrl.noaa.gov/gmd/ccgg/trends/data.html).
# This dataset measures atmospheric CO2 concentration at the Mauna Loa Observatory in
# Hawaii, and is widely used in the GP literature. It contains monthly CO2 readings
# starting in March 1958. Interestingly, there was an eruption at the Mauna Loa volcano in
# November 2022, so readings from December 2022 have changed to a site roughly 21 miles
# North of the Mauna Loa Observatory. We'll use the data from March 1958 to November 2022,
# and see how our GP extrapolates to 8 years before and after the data in the training
# set.
#
# First we'll load the data and plot it:

# %%
co2_data = pd.read_csv(
    "https://gml.noaa.gov/webdata/ccgg/trends/co2/co2_mm_mlo.csv", comment="#"
)
co2_data = co2_data.loc[co2_data["decimal date"] < 2022 + 11 / 12]
train_x = co2_data["decimal date"].values[:, None]
train_y = co2_data["average"].values[:, None]

fig, ax = plt.subplots()
ax.plot(train_x, train_y)
ax.set_title("CO2 Concentration in the Atmosphere")
ax.set_xlabel("Year")
ax.set_ylabel("CO2 Concentration (ppm)")

# %% [markdown]
# Looking at the data, we can see that there is clearly a periodic trend, with a period of
# roughly 1 year. We can also see that the data is increasing over time, which is
# also expected. This looks roughly linear, although it may have a non-linear component.
# This information will be useful when we come to choose our kernel.
#
# First, we'll construct our GPJax dataset, and will standardise the outputs, to match our
# assumption that the data has zero mean.

# %%
test_x = jnp.linspace(1950, 2030, 5000, dtype=jnp.float64).reshape(-1, 1)
y_scaler = StandardScaler().fit(train_y)
standardised_train_y = y_scaler.transform(train_y)

D = gpx.Dataset(X=train_x, y=standardised_train_y)

# %% [markdown]
# Having constructed our dataset, we'll now define our kernel. We'll use a kernel which is
# composed of the sum of a linear kernel and a periodic kernel, as we saw in the previous
# section that this kernel is able to capture both the periodic and linear trends in the
# data. We'll also add an RBF kernel to the sum, which will allow us to capture any
# non-linear trends in the data:
#
# $$\text{Kernel = Linear + Periodic + RBF}$$
#
#

# %%
mean = gpx.mean_functions.Zero()
rbf_kernel = gpx.kernels.RBF(lengthscale=100.0)
periodic_kernel = gpx.kernels.Periodic()
linear_kernel = gpx.kernels.Linear(variance=0.001)
sum_kernel = gpx.kernels.SumKernel(kernels=[linear_kernel, periodic_kernel])
final_kernel = gpx.kernels.SumKernel(kernels=[rbf_kernel, sum_kernel])

prior = gpx.gps.Prior(mean_function=mean, kernel=final_kernel)
likelihood = gpx.likelihoods.Gaussian(num_datapoints=D.n)

posterior = prior * likelihood


# %% [markdown]
# With our model constructed, let's now fit it to the data, by minimising the negative log
# marginal likelihood of the data:


# %%
def loss(posterior, data):
    return -gpx.objectives.conjugate_mll(posterior, data)


opt_posterior, history = gpx.fit(
    model=posterior,
    objective=loss,
    train_data=D,
    optim=ox.adamw(learning_rate=1e-2),
    num_iters=500,
    key=key,
)


# %% [markdown]
# Now we can obtain the model's prediction over a period of time which includes the
# training data, as well as 8 years before and after the training data:

# %%
latent_dist = opt_posterior.predict(test_x, train_data=D)
predictive_dist = opt_posterior.likelihood(latent_dist)

predictive_mean = predictive_dist.mean.reshape(-1, 1)
predictive_std = jnp.sqrt(predictive_dist.variance).reshape(-1, 1)

# %% [markdown]
# Let's plot the model's predictions over this period of time:

# %%
fig, ax = plt.subplots(figsize=(10, 5))
ax.plot(
    train_x, standardised_train_y, "x", label="Observations", color=cols[0], alpha=0.5
)
ax.fill_between(
    test_x.squeeze(),
    predictive_mean.squeeze() - 2 * predictive_std.squeeze(),
    predictive_mean.squeeze() + 2 * predictive_std.squeeze(),
    alpha=0.2,
    label="Two sigma",
    color=cols[1],
)
ax.plot(
    test_x,
    predictive_mean - 2 * predictive_std,
    linestyle="--",
    linewidth=1,
    color=cols[1],
)
ax.plot(
    test_x,
    predictive_mean + 2 * predictive_std,
    linestyle="--",
    linewidth=1,
    color=cols[1],
)
ax.plot(test_x, predictive_mean, label="Predictive mean", color=cols[1])
ax.set_xlabel("Year")
ax.legend(loc="center left", bbox_to_anchor=(0.975, 0.5))

# %% [markdown]
# We can see that the model seems to have captured the periodic trend in the data, as well
# as the (roughly) linear trend. This enables our model to make reasonable seeming
# predictions over the 8 years before and after the training data. Let's zoom in on the
# period from 2010 onwards:
#

# %%
fig, ax = plt.subplots(figsize=(10, 5))
ax.plot(
    train_x[train_x >= 2010],
    standardised_train_y[train_x >= 2010],
    "x",
    label="Observations",
    color=cols[0],
    alpha=0.5,
)
ax.fill_between(
    test_x[test_x >= 2010].squeeze(),
    predictive_mean[test_x >= 2010] - 2 * predictive_std[test_x >= 2010],
    predictive_mean[test_x >= 2010] + 2 * predictive_std[test_x >= 2010],
    alpha=0.2,
    label="Two sigma",
    color=cols[1],
)
ax.plot(
    test_x[test_x >= 2010],
    predictive_mean[test_x >= 2010] - 2 * predictive_std[test_x >= 2010],
    linestyle="--",
    linewidth=1,
    color=cols[1],
)
ax.plot(
    test_x[test_x >= 2010],
    predictive_mean[test_x >= 2010] + 2 * predictive_std[test_x >= 2010],
    linestyle="--",
    linewidth=1,
    color=cols[1],
)
ax.plot(
    test_x[test_x >= 2010],
    predictive_mean[test_x >= 2010],
    label="Predictive mean",
    color=cols[1],
)
ax.set_xlabel("Year")
ax.legend(loc="center left", bbox_to_anchor=(0.975, 0.5))

# %% [markdown]
# This certainly looks like a reasonable fit to the data, with sensible extrapolation
# beyond the training data, which finishes in November 2022. Moreover, the learned
# parameters of the kernel are interpretable. Let's take a look at the learned period of the periodic kernel:

# %%
print(
    "Periodic Kernel Period:"
    f" {[i for i in opt_posterior.prior.kernel.kernels if isinstance(i, gpx.kernels.Periodic)][0].period}"
)

# %% [markdown]
# This tells us that the periodic trend learned has a period of $\approx 1$. This makes
# intuitive sense, as the unit of the input data is years, and we can see that the
# periodic trend tends to repeat itself roughly every year!

# %% [markdown]
# ## Defining Kernels on Non-Euclidean Spaces
#
# In this notebook, we have focused solely on kernels whose domain resides in Euclidean space. However, what if one wished to work with data whose domain is non-Euclidean? For instance, one may wish to work with graph-structured data, or data which lies on a manifold, or even strings. Fortunately, kernels exist for a wide variety of domains. Whilst this is beyond the scope of this notebook, feel free to checkout out our [notebook on graph kernels](https://docs.jaxgaussianprocesses.com/_examples/graph_kernels/) for an introduction on how to define the Matérn kernel on graph-structured data, and there are a wide variety of resources online for learning about defining kernels in other domains. In terms of open-source libraries, the [Geometric Kernels](https://github.com/GPflow/GeometricKernels) library could be a good place to start if you're interested in looking at how these kernels may be implemented, with the additional benefit that it is compatible with GPJax.

# %% [markdown]
# ## Further Reading
#
# Congratulations on making it this far! We hope that this guide has given you a good introduction to kernels and how they can be used in GPJax. If you're interested in learning more about kernels, we recommend the following resources, which have also been used as inspiration for this guide:
#
# - [Gaussian Processes for Machine Learning](http://www.gaussianprocess.org/gpml/chapters/RW.pdf) - Chapter 4 provides a comprehensive overview of kernels, diving deep into some of the technical details and also providing some kernels defined on non-Euclidean spaces such as strings.
# - David Duvenaud's [Kernel Cookbook](https://www.cs.toronto.edu/~duvenaud/cookbook/) is a great resource for learning about kernels, and also provides some information about some of the pitfalls people commonly encounter when using the Matérn family of kernels. His PhD thesis, [Automatic Model Construction with Gaussian Processes](https://www.cs.toronto.edu/~duvenaud/thesis.pdf), also provides some in-depth recipes for how one may incorporate their prior knowledge when constructing kernels.
# - Finally, please check out our [more advanced kernel guide](https://docs.jaxgaussianprocesses.com/_examples/constructing_new_kernels/), which details some more kernels available in GPJax as well as how one may combine kernels together to form more complex kernels.
#
# ## System Configuration

# %%
# %reload_ext watermark
# %watermark -n -u -v -iv -w -a 'Thomas Christie'



from jax import (
    grad,
    jit,
)
import jax.numpy as jnp
import numpy as np
import pytest

from gpjax.numpyro_extras import FillTriangularTransform


# Helper function to generate a test input vector for a given matrix size.
def generate_test_vector(n):
    """
    Generate a sequential vector of shape (n(n+1)/2,) with values [1, 2, ..., n(n+1)/2].
    """
    L = n * (n + 1) // 2
    return jnp.arange(1, L + 1, dtype=jnp.float32)


# ----------------- Unit tests using PyTest -----------------


@pytest.mark.parametrize("n", [1, 2, 3, 4])
def test_forward_inverse(n):
    """
    Test that for a range of input sizes the forward transform correctly fills
    an n x n lower triangular matrix and that the inverse recovers the original vector.
    """
    ft = FillTriangularTransform()
    vec = generate_test_vector(n)
    L = ft(vec)

    # Construct the expected n x n lower triangular matrix
    expected = jnp.zeros((n, n), dtype=vec.dtype)
    row, col = jnp.tril_indices(n)
    expected = expected.at[row, col].set(vec)

    np.testing.assert_allclose(L, expected, rtol=1e-6)

    # Check that the inverse recovers the original vector
    vec_rec = ft.inv(L)
    np.testing.assert_allclose(vec, vec_rec, rtol=1e-6)


@pytest.mark.parametrize("n", [1, 2, 3, 4])
def test_batched_forward_inverse(n):
    """
    Test that the transform correctly handles batched inputs.
    """
    ft = FillTriangularTransform()
    batch_size = 5
    vec = jnp.stack([generate_test_vector(n) for _ in range(batch_size)], axis=0)
    L = ft(vec)  # Expected shape: (batch_size, n, n)
    assert L.shape == (batch_size, n, n)

    vec_rec = ft.inv(L)  # Expected shape: (batch_size, n(n+1)/2)
    assert vec_rec.shape == (batch_size, n * (n + 1) // 2)
    np.testing.assert_allclose(vec, vec_rec, rtol=1e-6)


def test_jit_forward():
    """
    Test that the forward transformation works correctly when compiled with JIT.
    """
    ft = FillTriangularTransform()
    n = 3
    vec = generate_test_vector(n)

    jit_forward = jit(ft)
    L = ft(vec)
    L_jit = jit_forward(vec)
    np.testing.assert_allclose(L, L_jit, rtol=1e-6)


def test_jit_inverse():
    """
    Test that the inverse transformation works correctly when compiled with JIT.
    """
    ft = FillTriangularTransform()
    n = 3
    vec = generate_test_vector(n)
    L_mat = ft(vec)

    # Wrap the inverse call in a lambda to avoid hashing the unhashable _InverseTransform.
    jit_inverse = jit(lambda y: ft.inv(y))
    vec_rec = ft.inv(L_mat)
    vec_rec_jit = jit_inverse(L_mat)
    np.testing.assert_allclose(vec_rec, vec_rec_jit, rtol=1e-6)


def test_grad_forward():
    """
    Test that JAX gradients can be computed for the forward transform.
    We define a simple function that sums the output matrix.
    Since the forward transform is just a reordering, the gradient should be 1
    for every element in the input vector.
    """
    ft = FillTriangularTransform()
    n = 3
    vec = generate_test_vector(n)

    # Define a scalar function f(x) = sum(forward(x))
    f = lambda x: jnp.sum(ft(x))
    grad_f = grad(f)(vec)
    np.testing.assert_allclose(grad_f, jnp.ones_like(vec), rtol=1e-6)


def test_grad_inverse():
    """
    Test that gradients flow through the inverse transformation.
    Define a simple scalar function on the inverse such that g(y) = sum(inv(y)).
    The gradient with respect to y should be one on the lower triangular indices.
    """
    ft = FillTriangularTransform()
    n = 3
    vec = generate_test_vector(n)
    L = ft(vec)

    g = lambda y: jnp.sum(ft.inv(y))
    grad_g = grad(g)(L)

    # Construct the expected gradient matrix: zeros everywhere except ones on the lower triangle.
    grad_expected = jnp.zeros_like(L)
    row, col = jnp.tril_indices(n)
    grad_expected = grad_expected.at[row, col].set(1.0)
    np.testing.assert_allclose(grad_g, grad_expected, rtol=1e-6)


def test_invalid_dimension_error():
    """
    Test that the FillTriangularTransform correctly raises a ValueError when
    the last dimension doesn't equal n(n+1)/2 for some integer n.
    """
    ft = FillTriangularTransform()

    # Create vectors with invalid dimensions that aren't n(n+1)/2 for any integer n
    invalid_dims = [2, 4, 5, 7, 8, 11, 13, 14, 17, 19, 20]

    for dim in invalid_dims:
        vec = jnp.ones(dim)
        with pytest.raises(
            ValueError,
            match="Last dimension must equal n\\(n\\+1\\)/2 for some integer n\\.",
        ):
            ft(vec)

    # Verify that valid dimensions don't raise errors
    valid_dims = [1, 3, 6, 10, 15, 21]  # n(n+1)/2 for n=1,2,3,4,5,6

    for dim in valid_dims:
        vec = jnp.ones(dim)
        try:
            ft(vec)
        except ValueError:
            pytest.fail(
                f"FillTriangularTransform raised ValueError for valid dimension {dim}"
            )


def test_inverse_dimension_error():
    """
    Test that the FillTriangularTransform.inv correctly raises a ValueError when
    the input has less than two dimensions.
    """
    ft = FillTriangularTransform()

    # Create a one-dimensional array
    vec = jnp.ones(3)  # 1D array with 3 elements

    # Try to call inverse on the 1D array, should fail
    with pytest.raises(
        ValueError, match="Input to inverse must be at least two-dimensional."
    ):
        ft.inv(vec)


def test_inverse_non_square_error():
    """
    Test that the FillTriangularTransform.inv correctly raises a ValueError when
    the input matrix is not square.
    """
    ft = FillTriangularTransform()

    # Create non-square matrices of different shapes
    non_square_matrices = [
        jnp.ones((3, 4)),  # 3x4 matrix
        jnp.ones((5, 2)),  # 5x2 matrix
        jnp.ones((1, 3)),  # 1x3 matrix
    ]

    for matrix in non_square_matrices:
        # Extract dimensions
        dim1, dim2 = matrix.shape[-2:]
        # Use a simpler regex pattern that doesn't include parentheses
        error_pattern = "Input matrix must be square; got shape"
        with pytest.raises(ValueError, match=error_pattern):
            ft.inv(matrix)

    # Test with batched non-square matrices
    batched_non_square = jnp.ones((2, 3, 4))  # Batch of 2 matrices of shape 3x4
    with pytest.raises(ValueError, match="Input matrix must be square"):
        ft.inv(batched_non_square)



from typing import TypeVar

from cola.annotations import PSD
from cola.linalg.decompositions.decompositions import Cholesky
from cola.linalg.inverse.inv import (
    inv,
    solve,
)
from cola.linalg.trace.diag_trace import diag
from cola.ops.operators import I_like
from flax import nnx
from jax import vmap
import jax.numpy as jnp
import jax.scipy as jsp
from jaxtyping import Float
import numpyro.distributions as npd
import typing_extensions as tpe

from gpjax.dataset import Dataset
from gpjax.distributions import GaussianDistribution
from gpjax.gps import (
    ConjugatePosterior,
    NonConjugatePosterior,
)
from gpjax.lower_cholesky import lower_cholesky
from gpjax.typing import (
    Array,
    ScalarFloat,
)
from gpjax.variational_families import AbstractVariationalFamily

VF = TypeVar("VF", bound=AbstractVariationalFamily)


Objective = tpe.Callable[[nnx.Module, Dataset], ScalarFloat]


def conjugate_mll(posterior: ConjugatePosterior, data: Dataset) -> ScalarFloat:
    r"""Evaluate the marginal log-likelihood of the Gaussian process.

    Compute the marginal log-likelihood function of the Gaussian process.
    The returned function can then be used for gradient based optimisation
    of the model's parameters or for model comparison. The implementation
    given here enables exact estimation of the Gaussian process' latent
    function values.

    For a training dataset $\{x_n, y_n\}_{n=1}^N$, set of test inputs
    $\mathbf{x}^{\star}$ the corresponding latent function evaluations are given
    by $\mathbf{f}=f(\mathbf{x})$ and $\mathbf{f}^{\star}f(\mathbf{x}^{\star})$,
    the marginal log-likelihood is given by:

    ```math
    \begin{align}
        \log p(\mathbf{y}) & = \int p(\mathbf{y}\mid\mathbf{f})
        p(\mathbf{f}, \mathbf{f}^{\star})\mathrm{d}\mathbf{f}^{\star}\\
        & = 0.5\left(-\mathbf{y}^{\top}\left(k(\mathbf{x}, \mathbf{x}')
        + \sigma^2\mathbf{I}_N\right)^{-1}\mathbf{y} \right.\\
        & \quad\left. -\log\lvert k(\mathbf{x}, \mathbf{x}')
        + \sigma^2\mathbf{I}_N\rvert - n\log 2\pi \right).
    \end{align}
    ```

    Example:
        >>> import gpjax as gpx

        >>> xtrain = jnp.linspace(0, 1).reshape(-1, 1)
        >>> ytrain = jnp.sin(xtrain)
        >>> D = gpx.Dataset(X=xtrain, y=ytrain)

        >>> meanf = gpx.mean_functions.Constant()
        >>> kernel = gpx.kernels.RBF()
        >>> likelihood = gpx.likelihoods.Gaussian(num_datapoints=D.n)
        >>> prior = gpx.gps.Prior(mean_function = meanf, kernel=kernel)
        >>> posterior = prior * likelihood

        >>> gpx.objectives.conjugate_mll(posterior, D)

        Our goal is to maximise the marginal log-likelihood. Therefore, when optimising
        the model's parameters with respect to the parameters, we use the negative
        marginal log-likelihood. This can be realised through

        >>> nmll = lambda p, d: -gpx.objectives.conjugate_mll(p, d)

    Args:
        posterior (ConjugatePosterior): The posterior distribution for which
            we want to compute the marginal log-likelihood.
        data:: The training dataset used to compute the
            marginal log-likelihood.

    Returns
    -------
        ScalarFloat: The marginal log-likelihood of the Gaussian process.
    """

    x, y = data.X, data.y

    # Observation noise o²
    obs_noise = posterior.likelihood.obs_stddev.value**2
    mx = posterior.prior.mean_function(x)

    # Σ = (Kxx + Io²) = LLᵀ
    Kxx = posterior.prior.kernel.gram(x)
    Kxx += I_like(Kxx) * posterior.prior.jitter
    Sigma = Kxx + I_like(Kxx) * obs_noise
    Sigma = PSD(Sigma)

    # p(y | x, θ), where θ are the model hyperparameters:
    mll = GaussianDistribution(jnp.atleast_1d(mx.squeeze()), Sigma)
    return mll.log_prob(jnp.atleast_1d(y.squeeze())).squeeze()


def conjugate_loocv(posterior: ConjugatePosterior, data: Dataset) -> ScalarFloat:
    r"""Evaluate the leave-one-out log predictive probability of the Gaussian process following
    section 5.4.2 of Rasmussen et al. 2006 - Gaussian Processes for Machine Learning. This metric
    calculates the average performance of all models that can be obtained by training on all but one
    data point, and then predicting the left out data point.

    The returned metric can then be used for gradient based optimisation
    of the model's parameters or for model comparison. The implementation
    given here enables exact estimation of the Gaussian process' latent
    function values.

    For a given ``ConjugatePosterior`` object, the following code snippet shows
    how the leave-one-out log predicitive probability can be evaluated.

    Example:
        >>> import gpjax as gpx
        ...
        >>> xtrain = jnp.linspace(0, 1).reshape(-1, 1)
        >>> ytrain = jnp.sin(xtrain)
        >>> D = gpx.Dataset(X=xtrain, y=ytrain)
        ...
        >>> meanf = gpx.mean_functions.Constant()
        >>> kernel = gpx.kernels.RBF()
        >>> likelihood = gpx.likelihoods.Gaussian(num_datapoints=D.n)
        >>> prior = gpx.gps.Prior(mean_function = meanf, kernel=kernel)
        >>> posterior = prior * likelihood
        ...
        >>> gpx.objectives.conjugate_loocv(posterior, D)

        Our goal is to maximise the leave-one-out log predictive probability. Therefore, when
        optimising the model's parameters with respect to the parameters, we use the negative
        leave-one-out log predictive probability. This can be realised through

        >>> nloocv = lambda p, d: -gpx.objectives.conjugate_loocv(p, d)

    Args:
        posterior (ConjugatePosterior): The posterior distribution for which
            we want to compute the marginal log-likelihood.
        data:: The training dataset used to compute the
            marginal log-likelihood.

    Returns
    -------
        ScalarFloat: The marginal log-likelihood of the Gaussian process.
    """

    x, y = data.X, data.y

    # Observation noise o²
    obs_var = posterior.likelihood.obs_stddev.value**2

    mx = posterior.prior.mean_function(x)  # [N, M]

    # Σ = (Kxx + Io²)
    Kxx = posterior.prior.kernel.gram(x)
    Sigma = Kxx + I_like(Kxx) * (obs_var + posterior.prior.jitter)
    Sigma = PSD(Sigma)  # [N, N]

    Sigma_inv_y = solve(Sigma, y - mx, Cholesky())  # [N, 1]
    Sigma_inv_diag = diag(inv(Sigma, Cholesky()))[:, None]  # [N, 1]

    loocv_means = mx + (y - mx) - Sigma_inv_y / Sigma_inv_diag
    loocv_stds = jnp.sqrt(1.0 / Sigma_inv_diag)

    loocv_posterior = npd.Normal(loc=loocv_means, scale=loocv_stds)
    return jnp.sum(loocv_posterior.log_prob(y))


def log_posterior_density(
    posterior: NonConjugatePosterior, data: Dataset
) -> ScalarFloat:
    r"""The log-posterior density of a non-conjugate Gaussian process. This is
    sometimes referred to as the marginal log-likelihood.

    Evaluate the log-posterior density of a Gaussian process.

    Compute the marginal log-likelihood, or log-posterior density of the Gaussian
    process. The returned function can then be used for gradient based optimisation
    of the model's parameters or for model comparison. The implementation given
    here is general and will work for any likelihood support by GPJax.

    Unlike the marginal_log_likelihood function of the `ConjugatePosterior` object,
    the marginal_log_likelihood function of the `NonConjugatePosterior` object does
    not provide an exact marginal log-likelihood function. Instead, the
    `NonConjugatePosterior` object represents the posterior distributions as a
    function of the model's hyperparameters and the latent function. Markov chain
    Monte Carlo, variational inference, or Laplace approximations can then be used
    to sample from, or optimise an approximation to, the posterior distribution.

    Args:
        posterior (NonConjugatePosterior): The posterior distribution for which
            we want to compute the marginal log-likelihood.
        data: The training dataset used to compute the
            marginal log-likelihood.

    Returns
    -------
        ScalarFloat: The log-posterior density of the Gaussian process.
    """

    x, y = data.X, data.y

    # Gram matrix
    Kxx = posterior.prior.kernel.gram(x)
    Kxx += I_like(Kxx) * posterior.prior.jitter
    Kxx = PSD(Kxx)
    Lx = lower_cholesky(Kxx)

    # Compute the prior mean function
    mx = posterior.prior.mean_function(x)

    # Whitened function values, wx, corresponding to the inputs, x
    wx = posterior.latent.value

    # f(x) = mx  +  Lx wx
    fx = mx + Lx @ wx

    # p(y | f(x), θ), where θ are the model hyperparameters
    likelihood = posterior.likelihood.link_function(fx)

    # Whitened latent function values prior, p(wx | θ) = N(0, I)
    latent_prior = npd.Normal(loc=0.0, scale=1.0)
    return likelihood.log_prob(y).sum() + latent_prior.log_prob(wx).sum()


non_conjugate_mll = log_posterior_density


def elbo(variational_family: VF, data: Dataset) -> ScalarFloat:
    r"""Compute the evidence lower bound of a variational approximation.

    Compute the evidence lower bound under this model. In short, this requires
    evaluating the expectation of the model's log-likelihood under the variational
    approximation. To this, we sum the KL divergence from the variational posterior
    to the prior. When batching occurs, the result is scaled by the batch size
    relative to the full dataset size.

    Args:
        variational_family: The variational
            approximation for whose parameters we should maximise the ELBO with
            respect to.
        data: The training data for which we should maximise the
            ELBO with respect to.

    Returns
    -------
        ScalarFloat: The evidence lower bound of the variational approximation.
    """
    # KL[q(f(·)) || p(f(·))]
    kl = variational_family.prior_kl()

    # ∫[log(p(y|f(·))) q(f(·))] df(·)
    var_exp = variational_expectation(variational_family, data)

    # For batch size b, we compute  n/b * Σᵢ[ ∫log(p(y|f(xᵢ))) q(f(xᵢ)) df(xᵢ)] - KL[q(f(·)) || p(f(·))]
    return (
        jnp.sum(var_exp)
        * variational_family.posterior.likelihood.num_datapoints
        / data.n
        - kl
    )


def variational_expectation(
    variational_family: VF,
    data: Dataset,
) -> Float[Array, " N"]:
    r"""Compute the variational expectation.

    Compute the expectation of our model's log-likelihood under our variational
    distribution. Batching can be done here to speed up computation.

    Args:
        variational_family: The variational family that we
            are using to approximate the posterior.
        data: The batch for which the expectation should be computed for.

    Returns
    -------
        Array: The expectation of the model's log-likelihood under our variational
            distribution.
    """
    # Unpack training batch
    x, y = data.X, data.y

    # Variational distribution q(f(·)) = N(f(·); μ(·), Σ(·, ·))
    q = variational_family

    # TODO: This needs cleaning up! We are squeezing then broadcasting `mean` and `variance`, which is not ideal.

    # Compute variational mean, μ(x), and variance, diag(Σ(x, x)), at the training
    # inputs, x
    def q_moments(x):
        qx = q(x)
        return qx.mean.squeeze(), qx.covariance().squeeze()

    mean, variance = vmap(q_moments)(x[:, None])

    # ≈ ∫[log(p(y|f(x))) q(f(x))] df(x)
    expectation = q.posterior.likelihood.expected_log_likelihood(
        y, mean[:, None], variance[:, None]
    )
    return expectation


# TODO: Replace code within CollapsedELBO to using (low rank structure of) LinOps and the GaussianDistribution object to be as succinct as e.g., the `ConjugateMLL`.


def collapsed_elbo(variational_family: VF, data: Dataset) -> ScalarFloat:
    r"""Compute a single step of the collapsed evidence lower bound.

    Compute the evidence lower bound under this model. In short, this requires
    evaluating the expectation of the model's log-likelihood under the variational
    approximation. To this, we sum the KL divergence from the variational posterior
    to the prior. When batching occurs, the result is scaled by the batch size
    relative to the full dataset size.

    Args:
        variational_family: The variational
            approximation for whose parameters we should maximise the ELBO with
            respect to.
        data: The training data for which we should maximise the
            ELBO with respect to.

    Returns
    -------
        ScalarFloat: The evidence lower bound of the variational approximation.
    """
    # Unpack training data
    x, y, n = data.X, data.y, data.n

    # Unpack mean function and kernel
    mean_function = variational_family.posterior.prior.mean_function
    kernel = variational_family.posterior.prior.kernel

    m = variational_family.num_inducing

    noise = variational_family.posterior.likelihood.obs_stddev.value**2
    z = variational_family.inducing_inputs.value
    Kzz = kernel.gram(z)
    Kzz += I_like(Kzz) * variational_family.jitter
    Kzz = PSD(Kzz)
    Kzx = kernel.cross_covariance(z, x)
    Kxx_diag = vmap(kernel, in_axes=(0, 0))(x, x)
    μx = mean_function(x)

    Lz = lower_cholesky(Kzz)

    # Notation and derivation:
    #
    # Let Q = KxzKzz⁻¹Kzx, we must compute the log normal pdf:
    #
    #   log N(y; μx, o²I + Q) = -nπ - n/2 log|o²I + Q|
    #   - 1/2 (y - μx)ᵀ (o²I + Q)⁻¹ (y - μx).
    #
    # The log determinant |o²I + Q| is computed via applying the matrix determinant
    #   lemma
    #
    #   |o²I + Q| = log|o²I| + log|I + Lz⁻¹ Kzx (o²I)⁻¹ Kxz Lz⁻¹| = log(o²) +  log|B|,
    #
    #   with B = I + AAᵀ and A = Lz⁻¹ Kzx / o.
    #
    # Similarly we apply matrix inversion lemma to invert o²I + Q
    #
    #   (o²I + Q)⁻¹ = (Io²)⁻¹ - (Io²)⁻¹ Kxz Lz⁻ᵀ (I + Lz⁻¹ Kzx (Io²)⁻¹ Kxz Lz⁻ᵀ )⁻¹ Lz⁻¹ Kzx (Io²)⁻¹
    #               = (Io²)⁻¹ - (Io²)⁻¹ oAᵀ (I + oA (Io²)⁻¹ oAᵀ)⁻¹ oA (Io²)⁻¹
    #               = I/o² - Aᵀ B⁻¹ A/o²,
    #
    # giving the quadratic term as
    #
    #   (y - μx)ᵀ (o²I + Q)⁻¹ (y - μx) = [(y - μx)ᵀ(y - µx)  - (y - μx)ᵀ Aᵀ B⁻¹ A (y - μx)]/o²,
    #
    #   with A and B defined as above.

    A = solve(Lz, Kzx, Cholesky()) / jnp.sqrt(noise)

    # AAᵀ
    AAT = jnp.matmul(A, A.T)

    # B = I + AAᵀ
    B = jnp.eye(m) + AAT

    # LLᵀ = I + AAᵀ
    L = jnp.linalg.cholesky(B)

    # log|B| = 2 trace(log|L|) = 2 Σᵢ log Lᵢᵢ  [since |B| = |LLᵀ| = |L|²  => log|B| = 2 log|L|, and |L| = Πᵢ Lᵢᵢ]
    log_det_B = 2.0 * jnp.sum(jnp.log(jnp.diagonal(L)))

    diff = y - μx

    # L⁻¹ A (y - μx)
    L_inv_A_diff = jsp.linalg.solve_triangular(L, jnp.matmul(A, diff), lower=True)

    # (y - μx)ᵀ (Io² + Q)⁻¹ (y - μx)
    quad = (jnp.sum(diff**2) - jnp.sum(L_inv_A_diff**2)) / noise

    # 2 * log N(y; μx, Io² + Q)
    two_log_prob = -n * jnp.log(2.0 * jnp.pi * noise) - log_det_B - quad

    # 1/o² tr(Kxx - Q) [Trace law tr(AB) = tr(BA) => tr(KxzKzz⁻¹Kzx) = tr(KxzLz⁻ᵀLz⁻¹Kzx) = tr(Lz⁻¹Kzx KxzLz⁻ᵀ) = trace(o²AAᵀ)]
    two_trace = jnp.sum(Kxx_diag) / noise - jnp.trace(AAT)

    # log N(y; μx, Io² + KxzKzz⁻¹Kzx) - 1/2o² tr(Kxx - KxzKzz⁻¹Kzx)
    return (two_log_prob - two_trace).squeeze() / 2.0



# Copyright 2023 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import typing as tp

import jax
import jax.numpy as jnp
import jax.random as jr
import optax as ox
from flax import nnx
from jax.flatten_util import ravel_pytree
from numpyro.distributions.transforms import Transform
from scipy.optimize import minimize

from gpjax.dataset import Dataset
from gpjax.objectives import Objective
from gpjax.parameters import (
    DEFAULT_BIJECTION,
    Parameter,
    transform,
)
from gpjax.scan import vscan
from gpjax.typing import (
    Array,
    KeyArray,
    ScalarFloat,
)

Model = tp.TypeVar("Model", bound=nnx.Module)


def fit(  # noqa: PLR0913
    *,
    model: Model,
    objective: Objective,
    train_data: Dataset,
    optim: ox.GradientTransformation,
    params_bijection: tp.Union[dict[Parameter, Transform], None] = DEFAULT_BIJECTION,
    key: KeyArray = jr.PRNGKey(42),
    num_iters: int = 100,
    batch_size: int = -1,
    log_rate: int = 10,
    verbose: bool = True,
    unroll: int = 1,
    safe: bool = True,
) -> tuple[Model, jax.Array]:
    r"""Train a Module model with respect to a supplied objective function.
    Optimisers used here should originate from Optax.

    Example:
    ```pycon
        >>> import jax.numpy as jnp
        >>> import jax.random as jr
        >>> import optax as ox
        >>> import gpjax as gpx
        >>> from gpjax.parameters import PositiveReal, Static
        >>>
        >>> # (1) Create a dataset:
        >>> X = jnp.linspace(0.0, 10.0, 100)[:, None]
        >>> y = 2.0 * X + 1.0 + 10 * jr.normal(jr.PRNGKey(0), X.shape)
        >>> D = gpx.Dataset(X, y)
        >>> # (2) Define your model:
        >>> class LinearModel(nnx.Module):
        >>>     def __init__(self, weight: float, bias: float):
        >>>         self.weight = PositiveReal(weight)
        >>>         self.bias = Static(bias)
        >>>
        >>>     def __call__(self, x):
        >>>         return self.weight.value * x + self.bias.value
        >>>
        >>> model = LinearModel(weight=1.0, bias=1.0)
        >>>
        >>> # (3) Define your loss function:
        >>> def mse(model, data):
        >>>     pred = model(data.X)
        >>>     return jnp.mean((pred - data.y) ** 2)
        >>>
        >>> # (4) Train!
        >>> trained_model, history = gpx.fit(
        >>>     model=model, objective=mse, train_data=D, optim=ox.sgd(0.001), num_iters=1000
        >>> )
    ```

    Args:
        model (Model): The model Module to be optimised.
        objective (Objective): The objective function that we are optimising with
            respect to.
        train_data (Dataset): The training data to be used for the optimisation.
        optim (GradientTransformation): The Optax optimiser that is to be used for
            learning a parameter set.
        num_iters (int): The number of optimisation steps to run. Defaults
            to 100.
        batch_size (int): The size of the mini-batch to use. Defaults to -1
            (i.e. full batch).
        key (KeyArray): The random key to use for the optimisation batch
            selection. Defaults to jr.PRNGKey(42).
        log_rate (int): How frequently the objective function's value should
            be printed. Defaults to 10.
        verbose (bool): Whether to print the training loading bar. Defaults
            to True.
        unroll (int): The number of unrolled steps to use for the optimisation.
            Defaults to 1.

    Returns:
        A tuple comprising the optimised model and training history.
    """
    if safe:
        # Check inputs.
        _check_model(model)
        _check_train_data(train_data)
        _check_optim(optim)
        _check_num_iters(num_iters)
        _check_batch_size(batch_size)
        _check_log_rate(log_rate)
        _check_verbose(verbose)

    # Model state filtering
    graphdef, params, *static_state = nnx.split(model, Parameter, ...)

    # Parameters bijection to unconstrained space
    if params_bijection is not None:
        params = transform(params, params_bijection, inverse=True)

    # Loss definition
    def loss(params: nnx.State, batch: Dataset) -> ScalarFloat:
        params = transform(params, params_bijection)
        model = nnx.merge(graphdef, params, *static_state)
        return objective(model, batch)

    # Initialise optimiser state.
    opt_state = optim.init(params)

    # Mini-batch random keys to scan over.
    iter_keys = jr.split(key, num_iters)

    # Optimisation step.
    def step(carry, key):
        params, opt_state = carry

        if batch_size != -1:
            batch = get_batch(train_data, batch_size, key)
        else:
            batch = train_data

        loss_val, loss_gradient = jax.value_and_grad(loss)(params, batch)
        updates, opt_state = optim.update(loss_gradient, opt_state, params)
        params = ox.apply_updates(params, updates)

        carry = params, opt_state
        return carry, loss_val

    # Optimisation scan.
    scan = vscan if verbose else jax.lax.scan

    # Optimisation loop.
    (params, _), history = scan(step, (params, opt_state), (iter_keys), unroll=unroll)

    # Parameters bijection to constrained space
    if params_bijection is not None:
        params = transform(params, params_bijection)

    # Reconstruct model
    model = nnx.merge(graphdef, params, *static_state)

    return model, history


def fit_scipy(  # noqa: PLR0913
    *,
    model: Model,
    objective: Objective,
    train_data: Dataset,
    max_iters: int = 500,
    verbose: bool = True,
    safe: bool = True,
) -> tuple[Model, Array]:
    r"""Train a Module model with respect to a supplied Objective function.
    Optimisers used here should originate from Optax. todo

    Args:
        model: the model Module to be optimised.
        objective: The objective function that we are optimising with
            respect to.
        train_data (Dataset): The training data to be used for the optimisation.
        max_iters (int): The maximum number of optimisation steps to run. Defaults
            to 500.
        verbose (bool): Whether to print the information about the optimisation. Defaults
            to True.

    Returns:
        A tuple comprising the optimised model and training history.
    """
    if safe:
        # Check inputs.
        _check_model(model)
        _check_train_data(train_data)
        _check_num_iters(max_iters)
        _check_verbose(verbose)

    # Model state filtering
    graphdef, params, *static_state = nnx.split(model, Parameter, ...)

    # Parameters bijection to unconstrained space
    params = transform(params, DEFAULT_BIJECTION, inverse=True)

    # Loss definition
    def loss(params) -> ScalarFloat:
        params = transform(params, DEFAULT_BIJECTION)
        model = nnx.merge(graphdef, params, *static_state)
        return objective(model, train_data)

    # convert to numpy for interface with scipy
    x0, scipy_to_jnp = ravel_pytree(params)

    @jax.jit
    def scipy_wrapper(x0):
        value, grads = jax.value_and_grad(loss)(scipy_to_jnp(jnp.array(x0)))
        scipy_grads = ravel_pytree(grads)[0]
        return value, scipy_grads

    history = [scipy_wrapper(x0)[0]]
    result = minimize(
        fun=scipy_wrapper,
        x0=x0,
        jac=True,
        callback=lambda X: history.append(scipy_wrapper(X)[0]),
        options={"maxiter": max_iters, "disp": verbose},
    )
    history = jnp.array(history)

    # convert back to nnx.State with JAX arrays
    params = scipy_to_jnp(result.x)

    # Parameters bijection to constrained space
    params = transform(params, DEFAULT_BIJECTION)

    # Reconstruct model
    model = nnx.merge(graphdef, params, *static_state)

    return model, history


def fit_lbfgs(
    *,
    model: Model,
    objective: Objective,
    train_data: Dataset,
    params_bijection: tp.Union[dict[Parameter, Transform], None] = DEFAULT_BIJECTION,
    max_iters: int = 100,
    safe: bool = True,
    max_linesearch_steps: int = 32,
    gtol: float = 1e-5,
) -> tuple[Model, jax.Array]:
    r"""Train a Module model with respect to a supplied Objective function.

    Uses Optax's LBFGS implementation and a jax.lax.while loop.

     Args:
         model: the model Module to be optimised.
         objective: The objective function that we are optimising with
             respect to.
         train_data (Dataset): The training data to be used for the optimisation.
         max_iters (int): The maximum number of optimisation steps to run. Defaults
             to 500.
         safe (bool): Whether to check the types of the inputs.
         max_linesearch_steps (int): The maximum number of linesearch steps to use
            for finding the stepsize.
        gtol (float): Terminate the optimisation if the L2 norm of the gradient is
            below this threshold.

     Returns:
         A tuple comprising the optimised model and final loss.
    """
    if safe:
        # Check inputs
        _check_model(model)
        _check_train_data(train_data)
        _check_num_iters(max_iters)

    # Model state filtering
    graphdef, params, *static_state = nnx.split(model, Parameter, ...)

    # Parameters bijection to unconstrained space
    if params_bijection is not None:
        params = transform(params, params_bijection, inverse=True)

    # Loss definition
    def loss(params: nnx.State) -> ScalarFloat:
        params = transform(params, params_bijection)
        model = nnx.merge(graphdef, params, *static_state)
        return objective(model, train_data)

    # Initialise optimiser
    optim = ox.lbfgs(
        linesearch=ox.scale_by_zoom_linesearch(
            max_linesearch_steps=max_linesearch_steps,
            initial_guess_strategy="one",
        )
    )
    opt_state = optim.init(params)
    loss_value_and_grad = ox.value_and_grad_from_state(loss)

    # Optimisation step.
    def step(carry):
        params, opt_state = carry

        # Using optax's value_and_grad_from_state is more efficient given LBFGS uses a linesearch
        # See https://optax.readthedocs.io/en/latest/api/utilities.html#optax.value_and_grad_from_state
        loss_val, loss_gradient = loss_value_and_grad(params, state=opt_state)
        updates, opt_state = optim.update(
            loss_gradient,
            opt_state,
            params,
            value=loss_val,
            grad=loss_gradient,
            value_fn=loss,
        )
        params = ox.apply_updates(params, updates)

        return params, opt_state

    def continue_fn(carry):
        _, opt_state = carry
        n = ox.tree_utils.tree_get(opt_state, "count")
        g = ox.tree_utils.tree_get(opt_state, "grad")
        g_l2_norm = ox.tree_utils.tree_l2_norm(g)
        return (n == 0) | ((n < max_iters) & (g_l2_norm >= gtol))

    # Optimisation loop
    params, opt_state = jax.lax.while_loop(
        continue_fn,
        step,
        (params, opt_state),
    )
    final_loss = ox.tree_utils.tree_get(opt_state, "value")

    # Parameters bijection to constrained space
    if params_bijection is not None:
        params = transform(params, params_bijection)

    # Reconstruct model
    model = nnx.merge(graphdef, params, *static_state)

    return model, final_loss


def get_batch(train_data: Dataset, batch_size: int, key: KeyArray) -> Dataset:
    """Batch the data into mini-batches. Sampling is done with replacement.

    Args:
        train_data (Dataset): The training dataset.
        batch_size (int): The batch size.
        key (KeyArray): The random key to use for the batch selection.

    Returns
    -------
        Dataset: The batched dataset.
    """
    x, y, n = train_data.X, train_data.y, train_data.n

    # Subsample mini-batch indices with replacement.
    indices = jr.choice(key, n, (batch_size,), replace=True)

    return Dataset(X=x[indices], y=y[indices])


def _check_model(model: tp.Any) -> None:
    """Check that the model is a subclass of nnx.Module."""
    if not isinstance(model, nnx.Module):
        raise TypeError(
            "Expected model to be a subclass of nnx.Module. "
            f"Got {model} of type {type(model)}."
        )


def _check_train_data(train_data: tp.Any) -> None:
    """Check that the train_data is of type gpjax.Dataset."""
    if not isinstance(train_data, Dataset):
        raise TypeError(
            "Expected train_data to be of type gpjax.Dataset. "
            f"Got {train_data} of type {type(train_data)}."
        )


def _check_optim(optim: tp.Any) -> None:
    """Check that the optimiser is of type GradientTransformation."""
    if not isinstance(optim, ox.GradientTransformation):
        raise TypeError(
            "Expected optim to be of type optax.GradientTransformation. "
            f"Got {optim} of type {type(optim)}."
        )


def _check_num_iters(num_iters: tp.Any) -> None:
    """Check that the number of iterations is of type int and positive."""
    if not isinstance(num_iters, int):
        raise TypeError(
            "Expected num_iters to be of type int. "
            f"Got {num_iters} of type {type(num_iters)}."
        )

    if num_iters <= 0:
        raise ValueError(f"Expected num_iters to be positive. Got {num_iters}.")


def _check_log_rate(log_rate: tp.Any) -> None:
    """Check that the log rate is of type int and positive."""
    if not isinstance(log_rate, int):
        raise TypeError(
            "Expected log_rate to be of type int. "
            f"Got {log_rate} of type {type(log_rate)}."
        )

    if not log_rate > 0:
        raise ValueError(f"Expected log_rate to be positive. Got {log_rate}.")


def _check_verbose(verbose: tp.Any) -> None:
    """Check that the verbose is of type bool."""
    if not isinstance(verbose, bool):
        raise TypeError(
            "Expected verbose to be of type bool. "
            f"Got {verbose} of type {type(verbose)}."
        )


def _check_batch_size(batch_size: tp.Any) -> None:
    """Check that the batch size is of type int and positive if not minus 1."""
    if not isinstance(batch_size, int):
        raise TypeError(
            "Expected batch_size to be of type int. "
            f"Got {batch_size} of type {type(batch_size)}."
        )

    if not batch_size == -1 and not batch_size > 0:
        raise ValueError(f"Expected batch_size to be positive or -1. Got {batch_size}.")


__all__ = [
    "fit",
    "get_batch",
]



import cola
from cola.ops.operators import (
    Dense,
    Diagonal,
)
import jax.numpy as jnp
import pytest

from gpjax.kernels.base import AbstractKernel
from gpjax.kernels.computations import (
    ConstantDiagonalKernelComputation,
    DiagonalKernelComputation,
)
from gpjax.kernels.nonstationary import (
    Linear,
    Polynomial,
)
from gpjax.kernels.stationary import (
    RBF,
    Matern12,
    Matern32,
    Matern52,
    Periodic,
    PoweredExponential,
    RationalQuadratic,
)


@pytest.mark.parametrize(
    "kernel",
    [
        RBF(),
        Matern12(),
        Matern32(),
        Matern52(),
        RationalQuadratic(),
        PoweredExponential(),
        Periodic(),
        Linear(),
        Polynomial(),
    ],
)
def test_change_computation(kernel: AbstractKernel):
    x = jnp.linspace(-3.0, 3.0, 5).reshape(-1, 1)

    # The default computation is DenseKernelComputation
    dense_linop = kernel.gram(x)
    dense_matrix = dense_linop.to_dense()
    dense_diagonals = jnp.diag(dense_matrix)

    assert isinstance(dense_linop, Dense)
    assert cola.PSD in dense_linop.annotations

    # Let's now change the computation to DiagonalKernelComputation
    kernel.compute_engine = DiagonalKernelComputation()
    diagonal_linop = kernel.gram(x)
    diagonal_matrix = diagonal_linop.to_dense()
    diag_entries = jnp.diag(diagonal_matrix)

    assert isinstance(diagonal_linop, Diagonal)
    assert cola.PSD in diagonal_linop.annotations

    # The diagonal entries should be the same as the dense matrix
    assert jnp.allclose(diag_entries, dense_diagonals)

    # All the off diagonal entries should be zero
    assert jnp.allclose(diagonal_matrix - jnp.diag(diag_entries), 0.0)

    # Let's now change the computation to ConstantDiagonalKernelComputation
    kernel.compute_engine = ConstantDiagonalKernelComputation()
    constant_diagonal_linop = kernel.gram(x)
    constant_diagonal_matrix = constant_diagonal_linop.to_dense()
    constant_entries = jnp.diag(constant_diagonal_matrix)

    assert cola.PSD in constant_diagonal_linop.annotations

    # Assert all the diagonal entries are the same
    assert jnp.allclose(constant_entries, constant_entries[0])

    # All the off diagonal entries should be zero
    assert jnp.allclose(constant_diagonal_matrix - jnp.diag(constant_entries), 0.0)



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import typing as tp

from flax import nnx
import jax.numpy as jnp
from jaxtyping import Float

from gpjax.kernels.computations import (
    AbstractKernelComputation,
    ConstantDiagonalKernelComputation,
)
from gpjax.kernels.stationary.base import StationaryKernel
from gpjax.typing import (
    Array,
    ScalarArray,
    ScalarFloat,
)


class White(StationaryKernel):
    r"""The White noise kernel.

    Computes the covariance for pairs of inputs $(x, y)$ with variance $\sigma^2$:
    $$
    k(x, y) = \sigma^2 \delta(x-y)
    $$
    """

    name: str = "White"

    def __init__(
        self,
        active_dims: tp.Union[list[int], slice, None] = None,
        variance: tp.Union[ScalarFloat, nnx.Variable[ScalarArray]] = 1.0,
        n_dims: tp.Union[int, None] = None,
        compute_engine: AbstractKernelComputation = ConstantDiagonalKernelComputation(),
    ):
        """Initializes the kernel.

        Args:
            active_dims: The indices of the input dimensions that the kernel operates on.
            variance: the variance of the kernel σ.
            n_dims: The number of input dimensions.
            compute_engine: The computation engine that the kernel uses to compute the
                covariance matrix
        """
        super().__init__(active_dims, 1.0, variance, n_dims, compute_engine)

    def __call__(self, x: Float[Array, " D"], y: Float[Array, " D"]) -> ScalarFloat:
        K = jnp.all(jnp.equal(x, y)) * self.variance.value
        return K.squeeze()



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import typing as tp

from cola.annotations import PSD
from cola.ops.operators import (
    Diagonal,
    Identity,
    Product,
)
from jax import vmap
import jax.numpy as jnp
from jaxtyping import Float

import gpjax
from gpjax.kernels.computations import AbstractKernelComputation
from gpjax.typing import Array

K = tp.TypeVar("K", bound="gpjax.kernels.base.AbstractKernel")  # noqa: F821
ConstantDiagonalType = Product


class ConstantDiagonalKernelComputation(AbstractKernelComputation):
    r"""Computation engine for constant diagonal kernels."""

    def gram(self, kernel: K, x: Float[Array, "N D"]) -> Product:
        value = kernel(x[0], x[0])
        dtype = value.dtype
        shape = (x.shape[0], x.shape[0])
        return PSD(jnp.atleast_1d(value) * Identity(shape=shape, dtype=dtype))

    def _diagonal(self, kernel: K, inputs: Float[Array, "N D"]) -> Diagonal:
        diag = vmap(lambda x: kernel(x, x))(inputs)
        return PSD(Diagonal(diag=diag))

    def _cross_covariance(
        self, kernel: K, x: Float[Array, "N D"], y: Float[Array, "M D"]
    ) -> Float[Array, "N M"]:
        # TODO: This is currently a dense implementation. We should implement
        # a sparse LinearOperator for non-square cross-covariance matrices.
        cross_cov = vmap(lambda x: vmap(lambda y: kernel(x, y))(y))(x)
        return cross_cov



# %% [markdown]
# Copyright 2022 The Jax Linear Operator Contributors All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from jax import config
import jax.numpy as jnp
import jax.random as jr
import pytest

# Enable Float64 for more stable matrix inversions.
config.update("jax_enable_x64", True)

import cola
from cola.ops import (
    Dense,
    Diagonal,
)

from gpjax.distributions import GaussianDistribution

_key = jr.key(seed=42)

from numpyro.distributions import MultivariateNormal
from numpyro.distributions.kl import kl_divergence


def approx_equal(res: jnp.ndarray, actual: jnp.ndarray) -> bool:
    """Check if two arrays are approximately equal."""
    return jnp.linalg.norm(res - actual) < 1e-5


@pytest.mark.parametrize("n", [1, 2, 5, 100])
def test_array_arguments(n: int) -> None:
    key_mean, key_sqrt = jr.split(_key, 2)
    mean = jr.uniform(key_mean, shape=(n,))
    sqrt = jr.uniform(key_sqrt, shape=(n, n))
    covariance = sqrt @ sqrt.T
    # check that cholesky does not error
    _L = jnp.linalg.cholesky(covariance)  # noqa: F841

    dist = GaussianDistribution(loc=mean, scale=cola.PSD(Dense(covariance)))

    assert approx_equal(dist.mean, mean)
    assert approx_equal(dist.variance, covariance.diagonal())
    assert approx_equal(dist.stddev(), jnp.sqrt(covariance.diagonal()))
    assert approx_equal(dist.covariance(), covariance)

    assert isinstance(dist.scale, Dense)
    assert cola.PSD in dist.scale.annotations

    y = jr.uniform(_key, shape=(n,))

    tfp_dist = MultivariateNormal(loc=mean, covariance_matrix=covariance)

    assert approx_equal(dist.log_prob(y), tfp_dist.log_prob(y))
    assert approx_equal(dist.kl_divergence(dist), 0.0)


@pytest.mark.parametrize("n", [1, 2, 5, 100])
def test_diag_linear_operator(n: int) -> None:
    key_mean, key_diag = jr.split(_key, 2)
    mean = jr.uniform(key_mean, shape=(n,))
    diag = jr.uniform(key_diag, shape=(n,))
    diag_covariance = jnp.diag(diag**2)

    # We purosely forget to add a PSD annotation to the diagonal matrix.
    dist_diag = GaussianDistribution(loc=mean, scale=Diagonal(diag**2))
    npt_dist = MultivariateNormal(loc=mean, covariance_matrix=diag_covariance)

    # We check that the PSD annotation is added automatically.
    assert isinstance(dist_diag.scale, Diagonal)
    assert cola.PSD in dist_diag.scale.annotations

    assert approx_equal(dist_diag.mean, npt_dist.mean)
    assert approx_equal(dist_diag.entropy(), npt_dist.entropy())
    assert approx_equal(dist_diag.variance, npt_dist.variance)
    assert approx_equal(dist_diag.covariance(), npt_dist.covariance_matrix)

    gpjax_samples = dist_diag.sample(key=_key, sample_shape=(10,))
    npt_samples = npt_dist.sample(key=_key, sample_shape=(10,))
    assert approx_equal(gpjax_samples, npt_samples)

    y = jr.uniform(_key, shape=(n,))

    assert approx_equal(dist_diag.log_prob(y), npt_dist.log_prob(y))

    assert approx_equal(dist_diag.kl_divergence(dist_diag), 0.0)


@pytest.mark.parametrize("n", [1, 2, 5, 100])
def test_dense_linear_operator(n: int) -> None:
    key_mean, key_sqrt = jr.split(_key, 2)
    mean = jr.uniform(key_mean, shape=(n,))
    sqrt = jr.uniform(key_sqrt, shape=(n, n))
    covariance = sqrt @ sqrt.T

    sqrt = jnp.linalg.cholesky(covariance + jnp.eye(n) * 1e-10)

    dist_dense = GaussianDistribution(loc=mean, scale=cola.PSD(Dense(covariance)))
    npt_dist = MultivariateNormal(loc=mean, covariance_matrix=covariance)

    assert approx_equal(dist_dense.mean, npt_dist.mean)
    assert approx_equal(dist_dense.entropy(), npt_dist.entropy())
    assert approx_equal(dist_dense.variance, npt_dist.variance)
    assert approx_equal(dist_dense.covariance(), npt_dist.covariance_matrix)

    y = jr.uniform(_key, shape=(n,))

    assert approx_equal(dist_dense.log_prob(y), npt_dist.log_prob(y))
    assert approx_equal(dist_dense.kl_divergence(dist_dense), 0.0)


@pytest.mark.parametrize("n", [1, 2, 5, 100])
def test_kl_divergence(n: int) -> None:
    key_a, key_b = jr.split(_key, 2)
    mean_a = jr.uniform(key_a, shape=(n,))
    mean_b = jr.uniform(key_b, shape=(n,))
    sqrt_a = jr.uniform(key_a, shape=(n, n))
    sqrt_b = jr.uniform(key_b, shape=(n, n))
    covariance_a = sqrt_a @ sqrt_a.T
    covariance_b = sqrt_b @ sqrt_b.T

    dist_a = GaussianDistribution(loc=mean_a, scale=cola.PSD(Dense(covariance_a)))
    dist_b = GaussianDistribution(loc=mean_b, scale=cola.PSD(Dense(covariance_b)))

    npt_dist_a = MultivariateNormal(loc=mean_a, covariance_matrix=covariance_a)
    npt_dist_b = MultivariateNormal(loc=mean_b, covariance_matrix=covariance_b)

    assert approx_equal(
        dist_a.kl_divergence(dist_b), kl_divergence(npt_dist_a, npt_dist_b)
    )
